/* 
 * Copyright (C) 2001-2007 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef SETTINGS_MANAGER_H
#define SETTINGS_MANAGER_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Util.h"
#include "Speaker.h"
#include "Singleton.h"

class SimpleXML;

class SettingsManagerListener {
public:
	virtual ~SettingsManagerListener() { }
	template<int I>	struct X { enum { TYPE = I };  };
	
	typedef X<0> Load;
	typedef X<1> Save;

	virtual void on(Load, SimpleXML&) throw() { }
	virtual void on(Save, SimpleXML&) throw() { }
};

class SettingsManager : public Singleton<SettingsManager>, public Speaker<SettingsManagerListener>
{
public:

	static StringList connectionSpeeds;
	static StringList connectionSpeeds1;

	enum StrSetting { STR_FIRST,
		CONNECTION = STR_FIRST, DESCRIPTION, DOWNLOAD_DIRECTORY, EMAIL, NICK, EXTERNAL_IP, 
		TEXT_FONT, MAINFRAME_ORDER, MAINFRAME_WIDTHS, HUBFRAME_ORDER, HUBFRAME_WIDTHS, 
		LANGUAGE_FILE, SEARCHFRAME_ORDER, SEARCHFRAME_WIDTHS, FAVORITESFRAME_ORDER, FAVORITESFRAME_WIDTHS, 
		HUBLIST_SERVERS, QUEUEFRAME_ORDER, QUEUEFRAME_WIDTHS, PUBLICHUBSFRAME_ORDER, PUBLICHUBSFRAME_WIDTHS, 
		USERSFRAME_ORDER, USERSFRAME_WIDTHS, HTTP_PROXY, LOG_DIRECTORY, LOG_FORMAT_POST_DOWNLOAD, 
		LOG_FORMAT_POST_UPLOAD, LOG_FORMAT_MAIN_CHAT, LOG_FORMAT_PRIVATE_CHAT, FINISHED_ORDER, FINISHED_WIDTHS, 
		TEMP_DOWNLOAD_DIRECTORY, BIND_ADDRESS, SOCKS_SERVER, SOCKS_USER, SOCKS_PASSWORD, CONFIG_VERSION,
		DEFAULT_AWAY_MESSAGE, ADLSEARCHFRAME_ORDER, ADLSEARCHFRAME_WIDTHS, SPYFRAME_WIDTHS, SPYFRAME_ORDER, 
		FINISHED_UL_WIDTHS, FINISHED_UL_ORDER, TIME_STAMPS_FORMAT, LOG_FILE_MAIN_CHAT, LOG_FILE_PRIVATE_CHAT, 
		LOG_FILE_STATUS, LOG_FILE_UPLOAD, LOG_FILE_DOWNLOAD, LOG_FILE_SYSTEM, LOG_FORMAT_SYSTEM, LOG_FORMAT_STATUS, 
		DIRECTORLISTINGFRAME_ORDER, DIRECTORLISTINGFRAME_WIDTHS, UPLOAD_SPEED, 
		BEEPFILE, WINAMP_FORMAT, TOOLBAR, EMOTICONFILE, 
		KICK_MSG_RECENT_01, KICK_MSG_RECENT_02, KICK_MSG_RECENT_03, KICK_MSG_RECENT_04, KICK_MSG_RECENT_05, 
		KICK_MSG_RECENT_06, KICK_MSG_RECENT_07, KICK_MSG_RECENT_08, KICK_MSG_RECENT_09, KICK_MSG_RECENT_10, 
		KICK_MSG_RECENT_11, KICK_MSG_RECENT_12, KICK_MSG_RECENT_13, KICK_MSG_RECENT_14, KICK_MSG_RECENT_15, 
		KICK_MSG_RECENT_16, KICK_MSG_RECENT_17, KICK_MSG_RECENT_18, KICK_MSG_RECENT_19, KICK_MSG_RECENT_20, 
		SEARCH_MSG_RECENT_01, SEARCH_MSG_RECENT_02, SEARCH_MSG_RECENT_03, SEARCH_MSG_RECENT_04, SEARCH_MSG_RECENT_05, 
		SEARCH_MSG_RECENT_06, SEARCH_MSG_RECENT_07, SEARCH_MSG_RECENT_08, SEARCH_MSG_RECENT_09, SEARCH_MSG_RECENT_10, 
		SEARCH_MSG_RECENT_11, SEARCH_MSG_RECENT_12, SEARCH_MSG_RECENT_13, SEARCH_MSG_RECENT_14, SEARCH_MSG_RECENT_15, 
		SEARCH_MSG_RECENT_16, SEARCH_MSG_RECENT_17, SEARCH_MSG_RECENT_18, SEARCH_MSG_RECENT_19, SEARCH_MSG_RECENT_20, 
		RECENTHUBSFRAME_ORDER, RECENTHUBSFRAME_WIDTHS, UPLOADQUEUEFRAME_ORDER, UPLOADQUEUEFRAME_WIDTHS, 
		MAINFRAME_VISIBLE, SEARCHFRAME_VISIBLE, QUEUEFRAME_VISIBLE, HUBFRAME_VISIBLE, UPLOADQUEUEFRAME_VISIBLE, 
		ANTIVIR_PATH, 
		STR_LAST };

	enum IntSetting { INT_FIRST = STR_LAST + 1,
		CONNECTION_TYPE = INT_FIRST, TCP_PORT, SLOTS, ROLLBACK, AUTO_FOLLOW, CLEAR_SEARCH, 
		BACKGROUND_COLOR, TEXT_COLOR, USE_OEM_MONOFONT, SHARE_HIDDEN, FILTER_MESSAGES, MINIMIZE_TRAY,
		AUTO_SEARCH, TIME_STAMPS, CONFIRM_EXIT, IGNORE_OFFLINE, POPUP_OFFLINE,
		LIST_DUPES, BUFFER_SIZE, DOWNLOAD_SLOTS, MAX_DOWNLOAD_SPEED, LOG_MAIN_CHAT, LOG_PRIVATE_CHAT,
		LOG_DOWNLOADS, LOG_UPLOADS, STATUS_IN_CHAT, SHOW_JOINS, PRIVATE_MESSAGE_BEEP, PRIVATE_MESSAGE_BEEP_OPEN,
		USE_SYSTEM_ICONS, POPUP_PMS, MIN_UPLOAD_SPEED, GET_USER_INFO, URL_HANDLER, MAIN_WINDOW_STATE,
		MAIN_WINDOW_SIZE_X, MAIN_WINDOW_SIZE_Y, MAIN_WINDOW_POS_X, MAIN_WINDOW_POS_Y, AUTO_AWAY,
		SOCKS_PORT, SOCKS_RESOLVE, KEEP_LISTS, AUTO_KICK, QUEUEFRAME_SHOW_TREE,
		COMPRESS_TRANSFERS, SHOW_PROGRESS_BARS, SFV_CHECK, MAX_TAB_ROWS, 
		MAX_COMPRESSION, ANTI_FRAG, MDI_MAXIMIZED, UDP_PORT, 
		SKIP_ZERO_BYTE, ADLS_BREAK_ON_FIRST, TAB_COMPLETION, 
		HUB_USER_COMMANDS, AUTO_SEARCH_AUTO_MATCH, UPLOAD_BAR_COLOR, DOWNLOAD_BAR_COLOR, LOG_SYSTEM, 
		LOG_FILELIST_TRANSFERS, SEND_UNKNOWN_COMMANDS, MAX_HASH_SPEED, SETTINGS_USE_CTRL_FOR_LINE_HISTORY, 
		GET_USER_COUNTRY, FAV_SHOW_JOINS, LOG_STATUS_MESSAGES, SHOW_STATUSBAR, 
		SHOW_TOOLBAR, SHOW_TRANSFERVIEW, POPUNDER_PM, POPUNDER_FILELIST, MAGNET_ASK, 
		MAGNET_ACTION, MAGNET_REGISTER, ADD_FINISHED_INSTANTLY, SETTINGS_USE_UPNP, DONT_DL_ALREADY_SHARED, 
		SEARCH_ONLY_TTH, TOGGLE_ACTIVE_WINDOW, PRIO_HIGHEST_SIZE, PRIO_HIGH_SIZE, 
		PRIO_NORMAL_SIZE, PRIO_LOW_SIZE, PRIO_LOWEST, MAX_FILELIST_SIZE, 
		OPEN_PUBLIC, OPEN_FAVORITE_HUBS, OPEN_FAVORITE_USERS, OPEN_QUEUE, OPEN_FINISHED_DOWNLOADS, 
		OPEN_FINISHED_UPLOADS, OPEN_SEARCH_SPY, OPEN_NOTEPAD, NO_IP_OVERRIDE, AUTO_KICK_NO_FAVS, PROMPT_PASSWORD,
		SEND_EXTENDED_INFO, GET_UPDATE_INFO, SEARCH_PASSIVE, SMALL_FILE_SIZE, SHUTDOWN_TIMEOUT, 
		MAX_UPLOAD_SPEED_LIMIT, FILTER_SEARCH, MAX_DOWNLOAD_SPEED_LIMIT, EXTRA_SLOTS, 
		TEXT_GENERAL_BACK_COLOR, TEXT_GENERAL_FORE_COLOR, TEXT_GENERAL_BOLD, TEXT_GENERAL_ITALIC, 
		TEXT_MYOWN_BACK_COLOR, TEXT_MYOWN_FORE_COLOR, TEXT_MYOWN_BOLD, TEXT_MYOWN_ITALIC, 
		TEXT_PRIVATE_BACK_COLOR, TEXT_PRIVATE_FORE_COLOR, TEXT_PRIVATE_BOLD, TEXT_PRIVATE_ITALIC, 
		TEXT_SYSTEM_BACK_COLOR, TEXT_SYSTEM_FORE_COLOR, TEXT_SYSTEM_BOLD, TEXT_SYSTEM_ITALIC, 
		TEXT_SERVER_BACK_COLOR, TEXT_SERVER_FORE_COLOR, TEXT_SERVER_BOLD, TEXT_SERVER_ITALIC, 
		TEXT_TIMESTAMP_BACK_COLOR, TEXT_TIMESTAMP_FORE_COLOR, TEXT_TIMESTAMP_BOLD, TEXT_TIMESTAMP_ITALIC, 
		TEXT_MYNICK_BACK_COLOR, TEXT_MYNICK_FORE_COLOR, TEXT_MYNICK_BOLD, TEXT_MYNICK_ITALIC, 
		TEXT_FAV_BACK_COLOR, TEXT_FAV_FORE_COLOR, TEXT_FAV_BOLD, TEXT_FAV_ITALIC, 
		TEXT_URL_BACK_COLOR, TEXT_URL_FORE_COLOR, TEXT_URL_BOLD, TEXT_URL_ITALIC, 
		BOLD_AUTHOR_MESS, MAX_UPLOAD_SPEED_LIMIT_NORMAL, THROTTLE_ENABLE, MAX_DOWNLOAD_SPEED_LIMIT_NORMAL, 
		MAX_UPLOAD_SPEED_LIMIT_TIME, MAX_DOWNLOAD_SPEED_LIMIT_TIME, TIME_DEPENDENT_THROTTLE, 
		BANDWIDTH_LIMIT_START, BANDWIDTH_LIMIT_END, PROGRESS_TEXT_COLOR_DOWN, 
		PROGRESS_TEXT_COLOR_UP, SHOW_INFOTIPS, EXTRA_DOWNLOAD_SLOTS, SOCKET_IN_BUFFER, SOCKET_OUT_BUFFER, 
		MINIMIZE_ON_STARTUP, FREE_SLOTS_DEFAULT, USE_EXTENSION_DOWNTO, 
		ERROR_COLOR, EXPAND_QUEUE, TRANSFER_SPLIT_SIZE, REPORT_ALTERNATES, AUTO_REFRESH_TIME, 
		SHOW_IP, I_DOWN_SPEED, H_DOWN_SPEED, DOWN_TIME, DISCONNECTING_ENABLE, MIN_FILE_SIZE, REMOVE_SLOW_USER, 
		SHOW_SHARE_CHEAT, PM_LOG_LINES, USE_EMOTICONS, MAX_EMOTICONS, ONLY_DL_TTH_FILES,
		TEXT_OP_BACK_COLOR, TEXT_OP_FORE_COLOR, TEXT_OP_BOLD, TEXT_OP_ITALIC, CLIENT_COMMANDS, 
		SERVER_COMMANDS, USER_COMMANDS, REPORT_DOWNLOAD, REPORT_UPLOAD, UPLOADQUEUEFRAME_SHOW_TREE, 
		GARBAGE_COMMAND_INCOMING, GARBAGE_COMMAND_OUTGOING, TTH_CHECK, CONFIRM_HUBEXIT, AWAY_ENABLED, 
		ALTERNATE_UPLOAD_SLOTS, TABS_ON_TOP, MUTE_SOUND, PAUSED_COLOR, USERLIST_DBLCLICK, TRANSFERLIST_DBLCLICK, 
		CHAT_DBLCLICK, USERLIST_COLORING, FAVORITE_COLOR, EXTRA_SLOT_COLOR, IGNORED_COLOR, 
		FIREBALL_COLOR, SERVER_COLOR, ACTIVE_COLOR, PASIVE_COLOR, OP_COLOR, NORMAL_COLOR, CONFIRM_HUB_RECONNECT, 
		LOG_PARTIAL_UPLOADS, OPEN_NETWORK_STATISTIC, OPEN_RECENT_HUBS, OPEN_UPLOAD_QUEUE, GROUP_SEARCH_RESULTS,
		SEARCH_EXPAND, FILELIST_DOWNLOAD_SLOTS, MAX_FILELIST2MATCH, AUTO_SEARCH_ON_AWAY, BOLD_FINISHED_DOWNLOADS, 
		BOLD_FINISHED_UPLOADS, BOLD_PM, JOIN_OPEN_NEW_WINDOW, BOLD_HUB, OPEN_SYSTEM_LOG, BOLD_QUEUE, 
		LAST_SEARCH_TYPE, SEARCH_ONLY_FREE_SLOTS, BOLD_SEARCH, BOLD_SYSTEM_LOG, CONFIRM_HUB_REMOVAL, 
		CONFIRM_ITEM_REMOVAL, USE_UPLOAD_SPEED, BOLD_WAITING_USERS, AUTO_SEARCH_LIMIT, ALT_SORT_ORDER, TTH_LISTS, 
		MAX_PMS, SPY_FRAME_IGNORE_TTH_SEARCHES, DONT_DL_ALREADY_QUEUED, MAX_COMMAND_LENGTH, FAST_HASH, 
		NO_SEG_UP, QUEUEFRAME_SPLITTER, WAITINGFRAME_SPLITTER, IP_IN_CHAT, ICONS_ON_TABS, MAX_TAB_TXT_LEN, 
		SORT_FAVUSERS_FIRST, SHOW_SHELL_MENU, FILTER_KICK_MESSAGES_NO_OP, 
		INT_LAST };

	enum Int64Setting { INT64_FIRST = INT_LAST + 1,
		TOTAL_UPLOAD = INT64_FIRST, TOTAL_DOWNLOAD, TOTAL_UPTIME, INT64_LAST, SETTINGS_LAST = INT64_LAST };

	enum { SPEED_MODEM, SPEED_ISDN, SPEED_SATELLITE, SPEED_WIRELESS, SPEED_CABLE,
		SPEED_DSL, SPEED_T1, SPEED_T3, SPEED_LAST };

	enum { CONNECTION_ACTIVE, CONNECTION_PASSIVE, CONNECTION_SOCKS5 };

	enum {	MAGNET_AUTO_SEARCH, MAGNET_AUTO_DOWNLOAD };

	const string& get(StrSetting key, bool useDefault = true) const {
		return (isSet[key] || !useDefault) ? strSettings[key - STR_FIRST] : strDefaults[key - STR_FIRST];
	}

	int get(IntSetting key, bool useDefault = true) const {
		return (isSet[key] || !useDefault) ? intSettings[key - INT_FIRST] : intDefaults[key - INT_FIRST];
	}
	int64_t get(Int64Setting key, bool useDefault = true) const {
		return (isSet[key] || !useDefault) ? int64Settings[key - INT64_FIRST] : int64Defaults[key - INT64_FIRST];
	}

	bool getBool(IntSetting key, bool useDefault = true) const {
		return (get(key, useDefault) != 0);
	}

	void set(StrSetting key, string const& value) {
		if((key == NICK) && (value.size() > 32)) {
			strSettings[key - STR_FIRST] = value.substr(0, 32);
		} else if((key == DESCRIPTION) && (value.size() > 64)) {
			strSettings[key - STR_FIRST] = value.substr(0, 64);
		} else if((key == EMAIL) && (value.size() > 64)) {
			strSettings[key - STR_FIRST] = value.substr(0, 64);
		} else {
			strSettings[key - STR_FIRST] = value;
		}
		isSet[key] = !value.empty();
	}

	int slots;
	void set(IntSetting key, int value) {
		if((key == SLOTS) && (value <= 0)) {
			value = 1;
		}
		if((key == SLOTS) ) {
			slots = value;
		}
		if((key == EXTRA_SLOTS) && (value < 3)) {
			value = 3;
		}

		if((key == SMALL_FILE_SIZE) && (value < 64)) {
			value = 64;
		}

		if((key == MAX_UPLOAD_SPEED_LIMIT_NORMAL) && (value > 0)) {
			if (value < ((2 * (slots)) + 3) ) {
				value = ((2 * (slots)) + 3);
			}
		}
		if((key == MAX_UPLOAD_SPEED_LIMIT_TIME) && (value > 0)) {
			if (value < ((2 * (slots)) + 3) ) {
				value = ((2 * (slots)) + 3);
			}
		}
		intSettings[key - INT_FIRST] = value;
		isSet[key] = true;
	}

	void set(IntSetting key, const string& value) {
		if(value.empty()) {
			intSettings[key - INT_FIRST] = 0;
			isSet[key] = false;
		} else {
			intSettings[key - INT_FIRST] = Util::toInt(value);
			isSet[key] = true;
		}
	}

	void set(Int64Setting key, int64_t value) {
		int64Settings[key - INT64_FIRST] = value;
		isSet[key] = true;
	}

	void set(Int64Setting key, const string& value) {
		if(value.empty()) {
			int64Settings[key - INT64_FIRST] = 0;
			isSet[key] = false;
		} else {
			int64Settings[key - INT64_FIRST] = Util::toInt64(value);
			isSet[key] = true;
		}
	}

	void set(IntSetting key, bool value) { set(key, (int)value); }

	void setDefault(StrSetting key, const string& value) {
		strDefaults[key - STR_FIRST] = value;
	}

	void setDefault(IntSetting key, int value) {
		intDefaults[key - INT_FIRST] = value;
	}
	void setDefault(Int64Setting key, int64_t value) {
		int64Defaults[key - INT64_FIRST] = value;
	}

	bool isDefault(int aSet) { return !isSet[aSet]; }

	void save() {
		save(Util::getConfigPath() + "CzDC.xml");
	}

	void load();
	void save(const string& aFileName);

private:
	friend class Singleton<SettingsManager>;
	SettingsManager();
	virtual ~SettingsManager() throw() { }

	static const string settingTags[SETTINGS_LAST+1];

	string strSettings[STR_LAST - STR_FIRST];
	int    intSettings[INT_LAST - INT_FIRST];
	int64_t int64Settings[INT64_LAST - INT64_FIRST];
	string strDefaults[STR_LAST - STR_FIRST];
	int    intDefaults[INT_LAST - INT_FIRST];
	int64_t int64Defaults[INT64_LAST - INT64_FIRST];
	bool isSet[SETTINGS_LAST];
};

// Shorthand accessor macros
#define SETTING(k) (SettingsManager::getInstance()->get(SettingsManager::k, true))
#define BOOLSETTING(k) (SettingsManager::getInstance()->getBool(SettingsManager::k, true))

#endif // SETTINGS_MANAGER_H

/* Some parts Copyright (C) 2002-2008 Petr Kozelka, PPK at czdc dot org */
