/* 
 * Copyright (C) 2001-2007 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include "../client/DCPlusPlus.h"
#include "Resource.h"

#include "HubFrame.h"
#include "LineDlg.h"
#include "SearchFrm.h"
#include "PrivateFrame.h"

#include "../client/QueueManager.h"
#include "../client/ShareManager.h"
#include "../client/Util.h"
#include "../client/FavoriteManager.h"
#include "../client/LogManager.h"
#include "../client/SettingsManager.h"
#include "../client/ConnectionManager.h" 
#include "../client/SearchManager.h"

HubFrame::FrameMap HubFrame::frames;

int HubFrame::columnSizes[] = { 100, 75, 75, 75, 100, 75, 100, 40, 40, 40, 40, 40, 40, 50 };
int HubFrame::columnIndexes[] = { COLUMN_NICK, COLUMN_SHARED, COLUMN_EXACT_SHARED, COLUMN_DESCRIPTION, COLUMN_TAG, 
	COLUMN_CONNECTION, COLUMN_EMAIL, COLUMN_CLIENTID, COLUMN_VERSION, COLUMN_MODE, COLUMN_HUBS, COLUMN_SLOTS, 
	COLUMN_LIMITER, COLUMN_IP };
static ResourceManager::Strings columnNames[] = { ResourceManager::NICK, ResourceManager::SHARED, ResourceManager::EXACT_SHARED, 
ResourceManager::DESCRIPTION, ResourceManager::TAG, ResourceManager::CONNECTION, ResourceManager::EMAIL, 
ResourceManager::CLIENTID, ResourceManager::VERSION, ResourceManager::MODE, ResourceManager::HUBS, ResourceManager::SLOTS, 
ResourceManager::LIMIT, ResourceManager::SETTINGS_IP };

tstring sSelectedURL = Util::emptyStringT;
long lURLBegin = 0;
long lURLEnd = 0;

LRESULT HubFrame::OnCreate(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& bHandled)
{
	CreateSimpleStatusBar(ATL_IDS_IDLEMESSAGE, WS_CHILD | WS_VISIBLE | WS_CLIPCHILDREN | WS_CLIPSIBLINGS | SBARS_SIZEGRIP);
	ctrlStatus.Attach(m_hWndStatusBar);

	ctrlClient.Create(m_hWnd, rcDefault, NULL, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | 
		WS_VSCROLL | ES_MULTILINE | ES_NOHIDESEL | ES_READONLY, WS_EX_CLIENTEDGE, IDC_CLIENT);

	ctrlClient.LimitText(0);
	ctrlClient.SetFont(WinUtil::font);
	clientContainer.SubclassWindow(ctrlClient.m_hWnd);
	
	ctrlClient.SetAutoURLDetect(false);
	ctrlClient.SetEventMask(ctrlClient.GetEventMask() | ENM_LINK);
	ctrlClient.SetClient(client);
	
	ctrlMessage.Create(m_hWnd, rcDefault, NULL, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | 
		ES_AUTOHSCROLL | ES_AUTOVSCROLL | ES_MULTILINE, WS_EX_CLIENTEDGE);
	
	ctrlMessageContainer.SubclassWindow(ctrlMessage.m_hWnd);
	ctrlMessage.SetFont(WinUtil::font);
	ctrlMessage.SetLimitText(9999);

	ctrlFilter.Create(m_hWnd, rcDefault, NULL, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | 
		ES_AUTOHSCROLL | WS_BORDER, WS_EX_CLIENTEDGE);

	ctrlFilterContainer.SubclassWindow(ctrlFilter.m_hWnd);
	ctrlFilter.SetFont(WinUtil::font);

	ctrlFilterSel.Create(m_hWnd, rcDefault, NULL, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_HSCROLL |
		WS_VSCROLL | CBS_DROPDOWNLIST | WS_BORDER, WS_EX_CLIENTEDGE);

	ctrlFilterSelContainer.SubclassWindow(ctrlFilterSel.m_hWnd);
	ctrlFilterSel.SetFont(WinUtil::font); 

	ctrlUsers.Create(m_hWnd, rcDefault, NULL, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | 
		WS_HSCROLL | WS_VSCROLL | LVS_REPORT | LVS_SHOWSELALWAYS | LVS_SHAREIMAGELISTS, WS_EX_CLIENTEDGE, IDC_USERS);

	if(Util::isXP) {
		ctrlUsers.SetExtendedListViewStyle(LVS_EX_LABELTIP | LVS_EX_HEADERDRAGDROP | LVS_EX_FULLROWSELECT | LVS_EX_INFOTIP);
	} else {
		ctrlUsers.SetExtendedListViewStyle(LVS_EX_LABELTIP | LVS_EX_HEADERDRAGDROP | LVS_EX_FULLROWSELECT | 0x00010000 | LVS_EX_INFOTIP);
	}

	splitChat.Create(m_hWnd);
	splitChat.SetSplitterPanes(ctrlClient.m_hWnd, ctrlMessage.m_hWnd, true);
	splitChat.SetSplitterExtendedStyle(0);
	splitChat.SetSplitterPos((WinUtil::fontHeight + 4)+5);

	SetSplitterPanes(splitChat.m_hWnd, ctrlUsers.m_hWnd, false);
	SetSplitterExtendedStyle(SPLIT_PROPORTIONAL);
	
	if(hubchatusersplit){ 
        m_nProportionalPos = hubchatusersplit;
	} else { 
        m_nProportionalPos = 7500;
    }
    
	ctrlShowUsers.Create(ctrlStatus.m_hWnd, rcDefault, _T("+/-"), WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN);
	ctrlShowUsers.SetButtonStyle(BS_AUTOCHECKBOX, false);
	ctrlShowUsers.SetFont(WinUtil::systemFont);
	ctrlShowUsers.SetCheck(ShowUserList ? BST_CHECKED : BST_UNCHECKED);
	showUsersContainer.SubclassWindow(ctrlShowUsers.m_hWnd);

    bool bColumsFromFavorite = false;
    if(sColumsOrder != Util::emptyString && sColumsWidth != Util::emptyString && sColumsVisible != Util::emptyString) bColumsFromFavorite = true;
    
    if(bColumsFromFavorite == false) {
		WinUtil::splitTokens(columnIndexes, SETTING(HUBFRAME_ORDER), COLUMN_LAST);
		WinUtil::splitTokens(columnSizes, SETTING(HUBFRAME_WIDTHS), COLUMN_LAST);                           
    } else {
		WinUtil::splitTokens(columnIndexes, sColumsOrder, COLUMN_LAST);
		WinUtil::splitTokens(columnSizes, sColumsWidth, COLUMN_LAST);
	}
    	
	for(int j=0; j < COLUMN_LAST; j++) {
		int fmt = (j == COLUMN_SHARED || j == COLUMN_EXACT_SHARED) ? LVCFMT_RIGHT : LVCFMT_LEFT;
		ctrlUsers.InsertColumn(j, CTSTRING_I(columnNames[j]), fmt, columnSizes[j], j);
	}
	
	ctrlUsers.setColumnOrderArray(COLUMN_LAST, columnIndexes);
    if(bColumsFromFavorite == true) {
	    ctrlUsers.setVisible(sColumsVisible);
    } else {
	    ctrlUsers.setVisible(SETTING(HUBFRAME_VISIBLE));
    }

	ctrlUsers.SetBkColor(WinUtil::bgColor);
	ctrlUsers.SetTextBkColor(WinUtil::bgColor);
	ctrlUsers.SetTextColor(WinUtil::textColor);
	ctrlUsers.setOwnerDraw(true);
	ctrlClient.SetBackgroundColor(WinUtil::bgColor); 
	
	ctrlUsers.setSortColumn(COLUMN_NICK);
				
	ctrlUsers.SetImageList(WinUtil::userImages, LVSIL_SMALL);

	CToolInfo ti(TTF_SUBCLASS, ctrlStatus.m_hWnd);
	
	ctrlLastLines.Create(ctrlStatus.m_hWnd, rcDefault, NULL, WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP, WS_EX_TOPMOST);
	ctrlLastLines.SetWindowPos(HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
	ctrlLastLines.AddTool(&ti);
	ctrlLastLines.SetDelayTime(TTDT_AUTOPOP, 30000);

	tabcopyMenu.CreatePopupMenu();
	tabcopyMenu.AppendMenu(MF_STRING, IDC_COPY_HUB, CTSTRING(HUB_NAME));
	tabcopyMenu.AppendMenu(MF_STRING, IDC_COPY_LINK, CTSTRING(HUB_ADDRESS));

	tabMenu.CreatePopupMenu();
	if(BOOLSETTING(LOG_MAIN_CHAT)) {
		tabMenu.AppendMenu(MF_STRING, IDC_OPEN_HUB_LOG, CTSTRING(OPEN_HUB_LOG));
		tabMenu.AppendMenu(MF_SEPARATOR);
	}
	tabMenu.AppendMenu(MF_STRING, IDC_ADD_AS_FAVORITE, CTSTRING(ADD_TO_FAVORITES));
	tabMenu.AppendMenu(MF_STRING, ID_FILE_RECONNECT, CTSTRING(MENU_RECONNECT));
	tabMenu.AppendMenu(MF_POPUP, (UINT)(HMENU)tabcopyMenu, CTSTRING(COPY));

	showJoins = BOOLSETTING(SHOW_JOINS);
	favShowJoins = BOOLSETTING(FAV_SHOW_JOINS);

	ctrlFilterSel.AddString(CTSTRING(NICK));
	ctrlFilterSel.AddString(CTSTRING(SHARED));
	ctrlFilterSel.AddString(CTSTRING(EXACT_SHARED));
	ctrlFilterSel.AddString(CTSTRING(DESCRIPTION));
	ctrlFilterSel.AddString(CTSTRING(TAG));
	ctrlFilterSel.AddString(CTSTRING(CONNECTION));
	ctrlFilterSel.AddString(CTSTRING(EMAIL));
	ctrlFilterSel.AddString(CTSTRING(CLIENTID));
	ctrlFilterSel.AddString(CTSTRING(VERSION));
	ctrlFilterSel.AddString(CTSTRING(MODE));
	ctrlFilterSel.AddString(CTSTRING(HUBS));
	ctrlFilterSel.AddString(CTSTRING(SLOTS));
	ctrlFilterSel.AddString(CTSTRING(LIMIT));
	ctrlFilterSel.AddString(CTSTRING(SETTINGS_IP));
	ctrlFilterSel.SetCurSel(0);

	bHandled = FALSE;
	client->connect();
	
	FavoriteManager::getInstance()->addListener(this);
	TimerManager::getInstance()->addListener(this);
	SettingsManager::getInstance()->addListener(this);
	
	return 1;
}

void HubFrame::openWindow(const tstring& aServer, 
		int windowposx, int windowposy, int windowsizex, int windowsizey, int windowtype, int chatusersplit, bool userliststate, 
        string sColumsOrder, string sColumsWidth, string sColumsVisible) {
	FrameIter i = frames.find(aServer);
	if(i == frames.end()) {
		HubFrame* frm = new HubFrame(aServer, chatusersplit, userliststate, sColumsOrder, sColumsWidth, sColumsVisible);
		frames[aServer] = frm;

		int nCmdShow = SW_SHOWDEFAULT;
		CRect rc = frm->rcDefault;

		rc.left = windowposx;
		rc.top = windowposy;
		rc.right = rc.left + windowsizex;
		rc.bottom = rc.top + windowsizey;
		if((rc.left < 0 ) || (rc.top < 0) || (rc.right - rc.left < 10) || ((rc.bottom - rc.top) < 10)) rc = frm->rcDefault;

		frm->CreateEx(WinUtil::mdiClient, rc);
		if(windowtype) frm->ShowWindow(((nCmdShow == SW_SHOWDEFAULT) || (nCmdShow == SW_SHOWNORMAL)) ? windowtype : nCmdShow);
	} else {
		if(::IsIconic(i->second->m_hWnd))
			::ShowWindow(i->second->m_hWnd, SW_RESTORE);
		i->second->MDIActivate(i->second->m_hWnd);
	}
}

LRESULT HubFrame::OnForwardMsg(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM lParam, BOOL& /*bHandled*/) {
	LPMSG pMsg = (LPMSG)lParam;
	if((pMsg->message >= WM_MOUSEFIRST) && (pMsg->message <= WM_MOUSELAST))
		ctrlLastLines.RelayEvent(pMsg);

	return 0;
}

void HubFrame::onEnter() {
	if(ctrlMessage.GetWindowTextLength() > 0) {
		AutoArray<TCHAR> msg(ctrlMessage.GetWindowTextLength()+1);
		ctrlMessage.GetWindowText(msg, ctrlMessage.GetWindowTextLength()+1);
		tstring s(msg, ctrlMessage.GetWindowTextLength());

		// save command in history, reset current buffer pointer to the newest command
		curCommandPosition = prevCommands.size();		//this places it one position beyond a legal subscript
		if(!curCommandPosition || curCommandPosition > 0 && prevCommands[curCommandPosition - 1] != s) {
			++curCommandPosition;
			prevCommands.push_back(s);
		}
		currentCommand = _T("");

		// Special command
		if(s[0] == _T('/')) {
			tstring m = s;
			tstring param;
			tstring message;
			tstring status;

			if(BOOLSETTING(CLIENT_COMMANDS)) {
				addClientLine(_T("Client command: ") + s, WinUtil::m_ChatTextSystem, true);
			}

			if(WinUtil::checkCommand(s, param, message, status)) {
				if(!message.empty()) {
					client->hubMessage(Text::fromT(message));
				}
				if(!status.empty()) {
					addClientLine(status, WinUtil::m_ChatTextSystem, true);
				}
			} else if(Util::stricmp(s.c_str(), _T("join"))==0) {
				if(!param.empty()) {
					redirect = param;
					if(BOOLSETTING(JOIN_OPEN_NEW_WINDOW)) {
						HubFrame::openWindow(param);
					} else {
						BOOL whatever = FALSE;
						onFollow(0, 0, 0, whatever);
					}
				} else {
					addClientLine(TSTRING(SPECIFY_SERVER), WinUtil::m_ChatTextSystem, true);
				}
			} else if((Util::stricmp(s.c_str(), _T("clear")) == 0) || (Util::stricmp(s.c_str(), _T("cls")) == 0)) {
				ctrlClient.ClearChat();
			} else if(Util::stricmp(s.c_str(), _T("ts")) == 0) {
				timeStamps = !timeStamps;
				if(timeStamps) {
					addClientLine(TSTRING(TIMESTAMPS_ENABLED), WinUtil::m_ChatTextSystem, true);
				} else {
					addClientLine(TSTRING(TIMESTAMPS_DISABLED), WinUtil::m_ChatTextSystem, true);
				}
			} else if((Util::stricmp(s.c_str(), _T("password")) == 0) && waitingForPW) {
				client->setPassword(Text::fromT(param));
				client->password(Text::fromT(param));
				waitingForPW = false;
			} else if( Util::stricmp(s.c_str(), _T("showjoins")) == 0 ) {
				showJoins = !showJoins;
				if(showJoins) {
					addClientLine(TSTRING(JOIN_SHOWING_ON), WinUtil::m_ChatTextSystem, true);
				} else {
					addClientLine(TSTRING(JOIN_SHOWING_OFF), WinUtil::m_ChatTextSystem, true);
				}
			} else if( Util::stricmp(s.c_str(), _T("favshowjoins")) == 0 ) {
				favShowJoins = !favShowJoins;
				if(favShowJoins) {
					addClientLine(TSTRING(FAV_JOIN_SHOWING_ON), WinUtil::m_ChatTextSystem, true);
				} else {
					addClientLine(TSTRING(FAV_JOIN_SHOWING_OFF), WinUtil::m_ChatTextSystem, true);
				}
			} else if(Util::stricmp(s.c_str(), _T("close")) == 0) {
				PostMessage(WM_CLOSE);
			} else if(Util::stricmp(s.c_str(), _T("userlist")) == 0) {
				ctrlShowUsers.SetCheck(ShowUserList ? BST_UNCHECKED : BST_CHECKED);
			} else if(Util::stricmp(s.c_str(), _T("connection")) == 0) {
				addClientLine(Text::toT((STRING(IP) + client->getLocalIp() + ", " + 
				STRING(PORT) + 
				Util::toString(ConnectionManager::getInstance()->getPort()) + "/" + 
				Util::toString(SearchManager::getInstance()->getPort()))), 
				WinUtil::m_ChatTextSystem, true);
			} else if((Util::stricmp(s.c_str(), _T("favorite")) == 0) || (Util::stricmp(s.c_str(), _T("fav")) == 0)) {
				addAsFavorite();
			} else if((Util::stricmp(s.c_str(), _T("removefavorite")) == 0) || (Util::stricmp(s.c_str(), _T("removefav")) == 0)) {
				removeFavoriteHub();
			} else if(Util::stricmp(s.c_str(), _T("getlist")) == 0){
				if(param.empty() == false){
					UserInfo* ui = findUser(param);
					if(ui != NULL) {
						ui->getList();
					} else {
                        OnlineUser* ou = client->findUser(Text::fromT(param));
                        if(ou != NULL) {
                        	try {
                        		QueueManager::getInstance()->addList(ou->getUser(), QueueItem::FLAG_CLIENT_VIEW);
                        	} catch(const Exception& e) {
                        		LogManager::getInstance()->message(e.getError());
                        	}
                        }
                    }
				}
			} else if(Util::stricmp(s.c_str(), _T("extraslots")) == 0) {
				int j = Util::toInt(Text::fromT(param));
				if(j > 0) {
					SettingsManager::getInstance()->set(SettingsManager::EXTRA_SLOTS, j);
					addClientLine(TSTRING(EXTRA_SLOTS_SET), WinUtil::m_ChatTextSystem, true);
				} else {
					addClientLine(TSTRING(INVALID_NUMBER_OF_SLOTS), WinUtil::m_ChatTextSystem, true);
				}
			} else if(Util::stricmp(s.c_str(), _T("smallfilesize")) == 0) {
				int j = Util::toInt(Text::fromT(param));
				if(j >= 64) {
					SettingsManager::getInstance()->set(SettingsManager::SMALL_FILE_SIZE, j);
					addClientLine(TSTRING(SMALL_FILE_SIZE_SET), WinUtil::m_ChatTextSystem, true);
				} else {
					addClientLine(TSTRING(INVALID_SIZE), WinUtil::m_ChatTextSystem, true);
				}
			} else if(Util::stricmp(s.c_str(), _T("savequeue")) == 0) {
				QueueManager::getInstance()->saveQueue();
				addClientLine(_T("Queue saved."), WinUtil::m_ChatTextSystem, true);
			} else if(Util::stricmp(s.c_str(), _T("whois")) == 0) {
				WinUtil::openLink(_T("http://www.ripe.net/perl/whois?form_type=simple&full_query_string=&searchtext=") + Text::toT(Util::encodeURI(Text::fromT(param))));
			} else if(Util::stricmp(s.c_str(), _T("ignorelist"))==0) {
				tstring ignorelist = _T("Ignored users:");
				for(TStringHash::iterator i = ignoreList.begin(); i != ignoreList.end(); ++i)
					ignorelist += _T(" ") + *i;
				addSystemLine(ignorelist, WinUtil::m_ChatTextSystem);
			} else if(Util::stricmp(s.c_str(), _T("log")) == 0) {
				StringMap params;
				params["hubNI"] = client->getHubName();
				params["hubURL"] = client->getHubUrl();
				params["myNI"] = client->getMyNick(); 
				if(param.empty()) {
					WinUtil::openFile(Text::toT(Util::validateFileName(SETTING(LOG_DIRECTORY) + Util::formatParams(SETTING(LOG_FILE_MAIN_CHAT), params, true))));
				} else if(Util::stricmp(param.c_str(), _T("status")) == 0) {
					WinUtil::openFile(Text::toT(Util::validateFileName(SETTING(LOG_DIRECTORY) + Util::formatParams(SETTING(LOG_FILE_STATUS), params, true))));
				}
			} else if(Util::stricmp(s.c_str(), _T("help")) == 0) {
				addSystemLine(_T("*** ") + WinUtil::commands + _T(", /smallfilesize #, /extraslots #, /savequeue, /join <hub-ip>, /clear, /ts, /showjoins, /favshowjoins, /close, /userlist, /connection, /favorite, /pm <user> [message], /getlist <user>, /log <status, system, downloads, uploads>, /winamp, /whois [IP], /ignorelist, /favoritenick, /favoritepassword, /removefavorite"), WinUtil::m_ChatTextSystem);
			} else if(Util::stricmp(s.c_str(), _T("pm")) == 0) {
				string::size_type j = param.find(_T(' '));
				if(j != string::npos) {
					tstring nick = param.substr(0, j);
					UserInfo* ui = findUser(nick);
					if(ui != NULL) {
						if(param.size() > j + 1)
							PrivateFrame::openWindow(ui->user, param.substr(j+1));
						else
							PrivateFrame::openWindow(ui->user);
					} else {
                        OnlineUser* ou = client->findUser(Text::fromT(nick));
                        if(ou != NULL) {
    						if(param.size() > j + 1)
    							PrivateFrame::openWindow(ou->getUser(), param.substr(j+1));
    						else
    							PrivateFrame::openWindow(ou->getUser());
                        }
                    }
				} else if(!param.empty()) {
					UserInfo* ui = findUser(param);
					if(ui != NULL) {
						PrivateFrame::openWindow(ui->user);
					} else {
                        OnlineUser* ou = client->findUser(Text::fromT(param));
                        if(ou != NULL) {
    						PrivateFrame::openWindow(ou->getUser());
                        }
                    }
				}
			} else if(Util::stricmp(s.c_str(), _T("me")) == 0) {
				client->hubMessage(Text::fromT(m));
			} else if((Util::stricmp(s.c_str(), _T("favoritenick")) == 0) || (Util::stricmp(s.c_str(), _T("favnick")) == 0)) {
				FavoriteHubEntry* hub = FavoriteManager::getInstance()->getFavoriteHubEntry(Text::fromT(server));
				if(hub) {
					if(!param.empty()) {
						hub->setNick(Text::fromT(param));
					} else {
						hub->setNick(client->getMyNick());
					}
					FavoriteManager::getInstance()->save();
					addClientLine(TSTRING(NICK_FOR_FAVORITE), WinUtil::m_ChatTextSystem, true);
				} else {
					addClientLine(TSTRING(HUB_NOT_FAVORITE), WinUtil::m_ChatTextSystem, true);
				}
			} else if((Util::stricmp(s.c_str(), _T("favoritepassword")) == 0) || (Util::stricmp(s.c_str(), _T("favpass")) == 0)) {
				FavoriteHubEntry* hub = FavoriteManager::getInstance()->getFavoriteHubEntry(Text::fromT(server));
				if(hub) {
					if(!param.empty()) {
						hub->setPassword(Text::fromT(param));
					} else {
						hub->setPassword(client->getPassword());
					}
					FavoriteManager::getInstance()->save();
					addClientLine(TSTRING(PASS_FOR_FAVORITE), WinUtil::m_ChatTextSystem, true);
				} else {
					addClientLine(TSTRING(HUB_NOT_FAVORITE), WinUtil::m_ChatTextSystem, true);
				}
			} else if((Util::stricmp(s.c_str(), _T("stat")) == 0) || (Util::stricmp(s.c_str(), _T("stats")) == 0)) {
			    if(client->isOp())
					client->hubMessage(WinUtil::generateStats());
			    else
					addSystemLine(Text::toT(WinUtil::generateStats()), WinUtil::m_ChatTextSystem);
			} else {
				if (BOOLSETTING(SEND_UNKNOWN_COMMANDS)) {
					client->hubMessage(Text::fromT(m));
				} else {
					addClientLine(TSTRING(UNKNOWN_COMMAND) + s, WinUtil::m_ChatTextSystem, true);
				}
			}
			ctrlMessage.SetWindowText(_T(""));
		} else if(waitingForPW) {
			addClientLine(TSTRING(DONT_REMOVE_SLASH_PASSWORD), WinUtil::m_ChatTextSystem, true);
			ctrlMessage.SetWindowText(_T("/password "));
			ctrlMessage.SetFocus();
			ctrlMessage.SetSel(10, 10);
		} else {
			if(BOOLSETTING(SERVER_COMMANDS)) {
				if(s[0] == '!' || s[0] == '+' || s[0] == '-')
					addClientLine(_T("Server command: ") + s, WinUtil::m_ChatTextSystem, true);
			}

			client->hubMessage(Text::fromT(s));
			ctrlMessage.SetWindowText(_T(""));
		}
	} else {
		MessageBeep(MB_ICONEXCLAMATION);
	}
}

struct CompareItems {
	CompareItems(int aCol) : col(aCol) { }
	bool operator()(const UserInfo& a, const UserInfo& b) const {
		return UserInfo::compareItems(&a, &b, col) < 0;
	}
	const int col;
};

void HubFrame::addAsFavorite() {
	FavoriteHubEntry* Hub = FavoriteManager::getInstance()->getFavoriteHubEntry(client->getHubUrl());
	if(!Hub) {
		FavoriteHubEntry aEntry;
		TCHAR buf[256];
		this->GetWindowText(buf, 255);
		aEntry.setServer(Text::fromT(server));
		aEntry.setName(Text::fromT(buf));
		aEntry.setDescription(Text::fromT(buf));
		aEntry.setConnect(false);
		aEntry.setNick(client->getMyNick());
		aEntry.setPassword(client->getPassword());
		aEntry.setConnect(false);
		FavoriteManager::getInstance()->addFavorite(aEntry);
		addClientLine(TSTRING(FAVORITE_HUB_ADDED), WinUtil::m_ChatTextSystem, true);
	} else {
		addClientLine(TSTRING(FAVORITE_HUB_ALREADY_EXISTS), WinUtil::m_ChatTextSystem, true);
	}
}

void HubFrame::removeFavoriteHub() {
	FavoriteHubEntry* removeHub = FavoriteManager::getInstance()->getFavoriteHubEntry(client->getHubUrl());
	if(removeHub) {
		FavoriteManager::getInstance()->removeFavorite(removeHub);
		addClientLine(TSTRING(FAVORITE_HUB_REMOVED), WinUtil::m_ChatTextSystem, true);
	} else {
		addClientLine(TSTRING(FAVORITE_HUB_DOES_NOT_EXIST), WinUtil::m_ChatTextSystem, true);
	}
}

LRESULT HubFrame::onCopyUserInfo(WORD /*wNotifyCode*/, WORD wID, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
    if(client->isConnected()) {
		if(ChatCtrl::sSelectedUser != _T("")) {
			UserInfo* ui = findUser(ChatCtrl::sSelectedUser);
			if(ui != NULL) {
				switch(wID) {
					case IDC_COPY_NICK:
						WinUtil::setClipboard(ui->getText(COLUMN_NICK, NULL));
						return 0;
					case IDC_COPY_EXACT_SHARE:
						WinUtil::setClipboard(ui->getText(COLUMN_EXACT_SHARED, NULL));
						return 0;
					case IDC_COPY_DESCRIPTION:
						WinUtil::setClipboard(ui->getText(COLUMN_DESCRIPTION, NULL));
						return 0;
					case IDC_COPY_TAG:
						WinUtil::setClipboard(ui->getText(COLUMN_TAG, NULL));
						return 0;
					case IDC_COPY_EMAIL_ADDRESS:
						WinUtil::setClipboard(ui->getText(COLUMN_EMAIL, NULL));
						return 0;
					case IDC_COPY_IP:
						WinUtil::setClipboard(ui->getText(COLUMN_IP, NULL));
						return 0;
					default:
						return 0;
				}
			}
			
			OnlineUser* ou = client->findUser(Text::fromT(ChatCtrl::sSelectedUser));
			if(ou != NULL) {
				switch(wID) {
					case IDC_COPY_NICK:
						WinUtil::setClipboard(Text::toT(ou->getIdentity().getNick()));
						return 0;
					case IDC_COPY_EXACT_SHARE:
						WinUtil::setClipboard(Text::toT(Util::formatExactSize(ou->getIdentity().getBytesShared())));
						return 0;
					case IDC_COPY_DESCRIPTION:
						WinUtil::setClipboard(Text::toT(Text::toUtf8(ou->getIdentity().getDescription())));
						return 0;
					case IDC_COPY_TAG:
						WinUtil::setClipboard(Text::toT(Text::toUtf8(ou->getIdentity().getTag())));
						return 0;
					case IDC_COPY_EMAIL_ADDRESS:
						WinUtil::setClipboard(Text::toT(Text::toUtf8(ou->getIdentity().getEmail())));
						return 0;
					case IDC_COPY_IP:
						WinUtil::setClipboard(Text::toT(ou->getIdentity().getIp()));
						return 0;
					default:
						return 0;
				}
			}
		}
    }
	return 0;
}

LRESULT HubFrame::onDoubleClickUsers(int /*idCtrl*/, LPNMHDR pnmh, BOOL& /*bHandled*/) {
	NMITEMACTIVATE* item = (NMITEMACTIVATE*)pnmh;
	if(item->iItem != -1) {
	    switch(SETTING(USERLIST_DBLCLICK)) {
		    case 0:
		          ctrlUsers.getItemData(item->iItem)->getList();
		          break;
		    case 1: {
                tstring sUser = Text::toT(ctrlUsers.getItemData(item->iItem)->getNick()) + _T(": ");
                ctrlMessage.AppendText(sUser.c_str());
                ctrlMessage.SetFocus();
                break;
		    }    
		    case 2:
		          if(ctrlUsers.getItemData(item->iItem)->user != ClientManager::getInstance()->getMe())
                        ctrlUsers.getItemData(item->iItem)->pm();
		          break;
		    case 3:
		          ctrlUsers.getItemData(item->iItem)->matchQueue();
		          break;
		    case 4:
		          ctrlUsers.getItemData(item->iItem)->grant();
		          break;
		    case 5:
		          ctrlUsers.getItemData(item->iItem)->addFav();
		          break;
		}	
	}
	return 0;
}

LRESULT HubFrame::onEditCopy(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	HWND focus = GetFocus();
	if ( focus == ctrlClient.m_hWnd ) {
		ctrlClient.Copy();
	}
	return 0;
}

LRESULT HubFrame::onEditSelectAll(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	HWND focus = GetFocus();
	if ( focus == ctrlClient.m_hWnd ) {
		ctrlClient.SetSelAll();
	}
	return 0;
}

LRESULT HubFrame::onEditClearAll(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	HWND focus = GetFocus();
	if(focus == ctrlClient.m_hWnd) {
		ctrlClient.ClearChat();
	}
	return 0;
}

bool HubFrame::updateUser(const UserTask& u, bool &bResort) {  
	UserMapIter i = userMap.find(u.user);
		
	if(i == userMap.end()) {
        if(ShowUserList == false) {
            userMap.insert(make_pair(u.user, (UserInfo*)NULL));
            return true;
        }
		UserInfo* ui = new UserInfo(u, false);
		bResort = ui->update(u.identity, ctrlUsers.getSortColumn()) || bResort;
		userMap.insert(make_pair(u.user, ui));
		if(filterstring.empty() || filterUser(u.identity) == true) {
			ctrlUsers.insertItem(ui, WinUtil::getImage(u.identity));
		}
		return true;
	} else {
        if(ShowUserList == false) {
            return false;
        }
		if(filterstring.empty() || filterUser(u.identity) == true) {
			UserInfo* ui = i->second;
			if(i->second != NULL) {
                bResort = ui->update(u.identity, ctrlUsers.getSortColumn()) || bResort;
    			int pos = ctrlUsers.findItem(ui);
    			if(pos != -1) {
    				ctrlUsers.updateItem(pos);
    				ctrlUsers.SetItem(pos, 0, LVIF_IMAGE, NULL, WinUtil::getImage(u.identity), 0, 0, NULL);
    			}
            } else {
        		ui = new UserInfo(u, false);
        		bResort = ui->update(u.identity, ctrlUsers.getSortColumn()) || bResort;
        		i->second = ui;
        		if(filterstring.empty() || filterUser(u.identity) == true) {
        			ctrlUsers.insertItem(ui, WinUtil::getImage(u.identity));
        		}
            }
		} else {
            if(i->second != NULL) {
    			int pos = ctrlUsers.findItem(i->second);
    			if(pos != -1) {
    				ctrlUsers.SetItemState(pos, 0, LVIS_SELECTED);
    				ctrlUsers.DeleteItem(pos);
    			}
            } else {
        		i->second = new UserInfo(u);
            }
		}
		return false;
	}
}

void HubFrame::removeUser(const User::Ptr& aUser) {
	UserMapIter i = userMap.find(aUser);
	
	if(i == userMap.end()) {
		return;
	}

	if(i->second != NULL) {
        UserInfo* ui = i->second;
		int pos = ctrlUsers.findItem(ui);
		if(pos != -1) {
			ctrlUsers.SetItemState(pos, 0, LVIS_SELECTED);
			ctrlUsers.DeleteItem(pos);
		}
		delete ui;
    }

	userMap.erase(i);
}

UserInfo* HubFrame::findUser(const tstring& nick) {
	for(UserMapIter i = userMap.begin(); i != userMap.end(); ++i) {
		if(i->second != NULL && _tcscmp(i->second->getText(COLUMN_NICK, NULL), nick.c_str()) == NULL)
			return i->second;
	}
	return 0;
}

LRESULT HubFrame::onSpeaker(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/) {
	TaskQueue::List t;
	tasks.get(t);

	if(t.size() > 2) {
		ctrlUsers.SetRedraw(FALSE);
	}
	
	bool bResort = false;

	for(TaskQueue::Iter i = t.begin(); i != t.end(); ++i) {
		switch(i->first) {
			case UPDATE_USER: {
				updateUser(*static_cast<UserTask*>(i->second), bResort);
				break;
			}	
        	case UPDATE_USER_JOIN: {
        		UserTask& u = *static_cast<UserTask*>(i->second);
        		if(updateUser(u, bResort)) {
        			if(showJoins || (favShowJoins && FavoriteManager::getInstance()->isFavoriteUser(u.user))) {
        				addSystemLine(_T("*** ") + TSTRING(JOINS) + Text::toT(u.identity.getNick()), WinUtil::m_ChatTextSystem);
        			}
        			if(client->isOp() && BOOLSETTING(SHOW_SHARE_CHEAT)) {
        				int64_t bytesSharedInt64 = u.identity.getBytesShared();
        				if(bytesSharedInt64 > 0) {
        					string bytesShared = Util::toString(bytesSharedInt64);
        					bool divided = (bytesSharedInt64 % (1024*1024) == 0);
        					bool samenumbers = false;
        					static const char* sSameNumbers[] = { "000000", "111111", "222222", "333333", "444444", "555555", "666666", "777777", "888888", "999999" };
        					for(int i = 0; i < 10; i++) {
        						if(strstr(bytesShared.c_str(), sSameNumbers[i]) != 0) {
        							samenumbers = true;
        							break;
        						}
        					}
        					char buf[512];
        					if(samenumbers && divided) {
        						sprintf_s(buf, 512, "*** User %s %s %s bytes", u.identity.getNick().c_str(), (STRING(SAMENUMBERS_DIVIDED)).c_str(), bytesShared.c_str());
        					} else if(samenumbers) {
        						sprintf_s(buf, 512, "*** User %s %s %s bytes", u.identity.getNick().c_str(), (STRING(MANY_SAMENUMBERS)).c_str(), bytesShared.c_str());
        					} else if(divided) {
        						sprintf_s(buf, 512, "*** User %s %s %s bytes", u.identity.getNick().c_str(), (STRING(DIVIDED_MB)).c_str(), bytesShared.c_str());
        					}
        					if(samenumbers || divided) {
        						addSystemLine(Text::toT(buf), WinUtil::m_ChatTextSystem);
        					}
        				}
        			}
        		}
        		break;
        	}
        	case REMOVE_USER: {
        		UserTask& u = *static_cast<UserTask*>(i->second);
                removeUser(u.user);
        		  
        		if(showJoins || (favShowJoins && FavoriteManager::getInstance()->isFavoriteUser(u.user))) {
        			addSystemLine(Text::toT("*** " + STRING(PARTS) + u.identity.getNick()), WinUtil::m_ChatTextSystem);
        		}
        		break;
    		}
    		case CONNECTED: {
	    		addClientLine(TSTRING(CONNECTED), WinUtil::m_ChatTextServer, true);
	    		setTabColor(RGB(0, 255, 0));
	    		unsetIconState();
	    		break;
	    	}    
	        case DISCONNECTED: {
	    		clearUserList();
	    		setTabColor(RGB(255, 0, 0));
	    		setIconState();
	    		break;
	    	} 
	        case ADD_CHAT_LINE: {
	        	MsgTask& chat = *static_cast<MsgTask*>(i->second);
	            if(&chat.identity == NULL || !ignoreList.count(Text::toT(chat.identity.getNick())) ||
	              (chat.identity.isOp() && !client->isOp())) {
	                addLine(chat.identity, Text::toT(chat.msg), WinUtil::m_ChatTextGeneral);
	            }
	    		break;
	    	} 
	        case ADD_STATUS_LINE: {
	        	addClientLine(Text::toT(static_cast<StringTask*>(i->second)->str), WinUtil::m_ChatTextServer, true);
	    		break;
	    	} 
	        case ADD_SILENT_STATUS_LINE: {
	        	addClientLine(Text::toT(static_cast<StringTask*>(i->second)->str), WinUtil::m_ChatTextSystem, false);
	    		break;
	    	} 
	        case SET_WINDOW_TITLE: {
	    		SetWindowText(Text::toT(static_cast<StringTask*>(i->second)->str).c_str());
	    		break;
	    	} 
	        case STATS: {
            	size_t userCount = userMap.size();	
            	size_t ShownUsers = ctrlUsers.GetItemCount();
            	string sStatus;
				if(ShowUserList == true) {
                	if(ctrlUsers.GetSelectedCount() > 1) {
                        sStatus += Util::toString(ctrlUsers.GetSelectedCount()) + "/";
                        ctrlStatus.SetText(2, Text::toT(Util::formatBytes(ctrlUsers.forEachSelectedT(CountAvailable()).available) + "/" + 
                            Util::formatBytes(CountUserMapAvailable())).c_str());
                    } else {
                        ctrlStatus.SetText(2, Text::toT(Util::formatBytes(CountUserMapAvailable())).c_str());
                    }
                	if(filterstring.empty() == false && userCount != ShownUsers) {
                        sStatus += Util::toString(ShownUsers) + "/";
                	}
                } else {
					ctrlStatus.SetText(2, Text::toT(Util::formatBytes(client->getAvailable())).c_str());
                }
            	ctrlStatus.SetText(1, Text::toT(sStatus + Util::toString(userCount) + " " + STRING(HUB_USERS)).c_str());
	    		break;
	    	}    	
	        case GET_PASSWORD: {
	    		if(client->getPassword().size() > 0) {
	    			client->password(client->getPassword());
	    			addClientLine(TSTRING(STORED_PASSWORD_SENT), WinUtil::m_ChatTextSystem, true);
	    		} else {
					if(!BOOLSETTING(PROMPT_PASSWORD)) {
	    				ctrlMessage.SetWindowText(_T("/password "));
	    				ctrlMessage.SetFocus();
	    				ctrlMessage.SetSel(10, 10);
	    				waitingForPW = true;
					} else {
						LineDlg linePwd;
						linePwd.title = CTSTRING(ENTER_PASSWORD);
						linePwd.description = CTSTRING(ENTER_PASSWORD);
						linePwd.password = true;
						if(linePwd.DoModal(m_hWnd) == IDOK) {
							client->setPassword(Text::fromT(linePwd.line));
							client->password(Text::fromT(linePwd.line));
							waitingForPW = false;
						} else {
							client->disconnect(true);
						}
					}
	    		}
	    		break;
	    	}
	        case PRIVATE_MESSAGE: {
	        	PMTask& pm = *static_cast<PMTask*>(i->second);
	    		if(!ignoreList.count(Text::toT(pm.fromidentity.getNick())) || (pm.fromidentity.isOp() && !client->isOp())) {
					if(pm.fromidentity.getUser()->isOnline()) {
	    				if((BOOLSETTING(POPUP_PMS) && PrivateFrame::getPmCount() < SETTING(MAX_PMS)) || PrivateFrame::isOpen(pm.fromidentity.getUser())) {
	    					PrivateFrame::gotMessage(pm.fromidentity, pm.byidentity, Text::toT(pm.msg));
	    				} else {
							addSystemLine(TSTRING(PRIVATE_MESSAGE_FROM) + Text::toT(pm.fromidentity.getNick()) + _T(": ") + Text::toT(pm.msg), WinUtil::m_ChatTextPrivate);
	    				}
	    				if(BOOLSETTING(MINIMIZE_TRAY) && (pm.fromidentity.getConnection().empty() == false || pm.fromidentity.getBytesShared() != 0)) {
	    					HWND hMainWnd = GetTopLevelWindow();
	    					::PostMessage(hMainWnd, WM_SPEAKER, 7, NULL);
	    				}
	    			} else {
	    				if(BOOLSETTING(IGNORE_OFFLINE)) {
	    					addClientLine(TSTRING(IGNORED_MESSAGE) + Text::toT(pm.msg), WinUtil::m_ChatTextPrivate, false);
	    				} else if(BOOLSETTING(POPUP_OFFLINE) && PrivateFrame::getPmCount() < SETTING(MAX_PMS)) {
	    					PrivateFrame::gotMessage(pm.fromidentity, pm.byidentity, Text::toT(pm.msg));
	    				} else {
	    					addSystemLine(TSTRING(PRIVATE_MESSAGE_FROM) + Text::toT(pm.fromidentity.getNick()) + _T(": ") + Text::toT(pm.msg), WinUtil::m_ChatTextPrivate);
	    				}
					}
	    		}
	    		break;
	    	}
	        case KICK_MSG: {
	            MsgTask& kick = *static_cast<MsgTask*>(i->second);
	            if(BOOLSETTING(FILTER_MESSAGES) == true || (BOOLSETTING(FILTER_KICK_MESSAGES_NO_OP) == true && client->isOp() == false)) {
	                addClientLine(Text::toT(kick.msg), WinUtil::m_ChatTextSystem, false, true);
	            } else {
	        		addLine(kick.identity, Text::toT(kick.msg), WinUtil::m_ChatTextServer, false);
	            }
	            break;
	    	}
	    	case PRIVATE_MESSAGE_NICK: {
	    		PMNickTask& pm = *static_cast<PMNickTask*>(i->second);
	    		if(!ignoreList.count(Text::toT(pm.nick))) {
	    			if(BOOLSETTING(IGNORE_OFFLINE)) {
	    				addClientLine(TSTRING(IGNORED_MESSAGE) + Text::toT(pm.msg), WinUtil::m_ChatTextPrivate, false);
	    			} else if(BOOLSETTING(POPUP_OFFLINE)) {
						Identity id = Identity(ClientManager::getInstance()->getUser(pm.nick, client->getHubUrl()));
						PrivateFrame::gotMessage(id, pm.byidentity, Text::toT(pm.msg));
	    			} else {
	    				addSystemLine(TSTRING(PRIVATE_MESSAGE_FROM) + Text::toT(pm.nick) + _T(": ") + Text::toT(pm.msg), WinUtil::m_ChatTextPrivate);
					}
	    		}
	    	}
			case RESORT_USERLIST:
				bResort = true;
				break;
	    	default:
	            break;
    	}
		delete i->second;
	}
        
	if(bResort) {
    	ctrlUsers.resort();
    }
    
	if(t.size() > 2) {
    	ctrlUsers.SetRedraw(TRUE);
    }

	return 0;
}

void HubFrame::UpdateLayout(BOOL bResizeBars /* = TRUE */) {
	RECT rect;
	GetClientRect(&rect);
	// position bars and offset their dimensions
	UpdateBarsPosition(rect, bResizeBars);
	
	if(ctrlStatus.IsWindow()) {
		CRect sr;
		int w[4];
		ctrlStatus.GetClientRect(sr);

		int tmp = (sr.Width()) > 332 ? 232 : ((sr.Width() > 132) ? sr.Width()-100 : 32);
		
		w[0] = sr.right - tmp;
		w[1] = w[0] + (tmp-32)/2;
		w[2] = w[0] + (tmp-32);
		w[3] = w[2] + 16;
		
		ctrlStatus.SetParts(4, w);

		ctrlLastLines.SetMaxTipWidth(w[0]);

		// Strange, can't get the correct width of the last field...
		ctrlStatus.GetRect(2, sr);
		sr.left = sr.right + 2;
		sr.right = sr.left + 16;
		ctrlShowUsers.MoveWindow(sr);
	}
	int h = WinUtil::fontHeight + 4;

	CRect rc = rect;
	if(!ShowUserList) {
		if(GetSinglePaneMode() == SPLIT_PANE_NONE)
			SetSinglePaneMode(SPLIT_PANE_LEFT);
	} else {
		if(GetSinglePaneMode() != SPLIT_PANE_NONE)
			SetSinglePaneMode(SPLIT_PANE_NONE);
	}
	SetSplitterRect(rc);

	if(rc.Height() > ((2 * h) + 5)) {
		splitChat.SetSplitterPos( rc.Height() - ((2 * h) + 5) );
	} else {
		splitChat.SetSplitterPos(rc.Height()-1);
	}

	if(ShowUserList){
		GetSplitterPaneRect(1, &rc);
		rc.bottom = rect.bottom - h - 10;
		ctrlUsers.MoveWindow(rc);
		
		rc.top = rc.bottom + 3;
		rc.bottom = rect.bottom - 2;
		rc.right = rc.right - 100;
		ctrlFilter.MoveWindow(rc);

		rc.left = rc.right + 4;
		rc.right = rect.right;
		rc.bottom = rc.bottom + 120;
		ctrlFilterSel.MoveWindow(rc); 
	}
}

LRESULT HubFrame::onClose(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& bHandled) {
	if(!closed) {
		if(MainFrame::closing == true || !BOOLSETTING(CONFIRM_HUBEXIT) || (MessageBox(CTSTRING(CONFIRM_HUBCLOSE), _T(APPNAME) _T(" ") _T(VERSIONSTRING), MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2) == IDYES)) {
			RecentHubEntry* r = FavoriteManager::getInstance()->getRecentHubEntry(Text::fromT(server));
			if(r) {
				TCHAR buf[256];
				this->GetWindowText(buf, 255);
				r->setName(Text::fromT(buf));
				r->setUsers(Util::toString(client->getUserCount()));
				r->setShared(Util::toString(client->getAvailable()));
				FavoriteManager::getInstance()->updateRecent(r);
			}
			TimerManager::getInstance()->removeListener(this);
			FavoriteManager::getInstance()->removeListener(this);
			SettingsManager::getInstance()->removeListener(this);
			client->removeListener(this);
			client->disconnect(true);
	
			closed = true;
			PostMessage(WM_CLOSE);
		}
		return 0;
	} else {
		SettingsManager::getInstance()->set(SettingsManager::GET_USER_INFO, ShowUserList);
		FavoriteManager::getInstance()->removeUserCommand(Text::fromT(server));

		clearUserList();
		clearTaskList();
			
		FavoriteHubEntry* hub = FavoriteManager::getInstance()->getFavoriteHubEntry(Text::fromT(server));
		if(hub) {
			WINDOWPLACEMENT wp;
			wp.length = sizeof(wp);
			GetWindowPlacement(&wp);

			CRect rc;
			GetWindowRect(rc);
			CRect rcmdiClient;
			::GetWindowRect(WinUtil::mdiClient, &rcmdiClient);
			if(wp.showCmd == SW_SHOW || wp.showCmd == SW_SHOWNORMAL) {
				hub->setWindowPosX(rc.left - (rcmdiClient.left + 2));
				hub->setWindowPosY(rc.top - (rcmdiClient.top + 2));
				hub->setWindowSizeX(rc.Width());
				hub->setWindowSizeY(rc.Height());
			}
			if(wp.showCmd == SW_SHOWNORMAL || wp.showCmd == SW_SHOW || wp.showCmd == SW_SHOWMAXIMIZED || wp.showCmd == SW_MAXIMIZE)
				hub->setWindowType((int)wp.showCmd);
			hub->setChatUserSplit(m_nProportionalPos);
			hub->setUserListState(ShowUserList);
			ctrlUsers.saveFavoriteHeaderOrder(hub);
			FavoriteManager::getInstance()->save();
		} else {
			ctrlUsers.saveHeaderOrder(SettingsManager::HUBFRAME_ORDER, SettingsManager::HUBFRAME_WIDTHS,
				SettingsManager::HUBFRAME_VISIBLE);
		}
	
		bHandled = FALSE;
		return 0;
	}
}

void HubFrame::clearUserList() {
	ctrlUsers.DeleteAllItems();
	for(UserMapIter i = userMap.begin(); i != userMap.end(); ++i) {
		if(i->second != NULL) {
			delete i->second;
		}
	}
	userMap.clear();
}

void HubFrame::clearTaskList() {
	tasks.clear();
}

LRESULT HubFrame::onLButton(UINT /*uMsg*/, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	HWND focus = GetFocus();
	bHandled = false;
	if(focus == ctrlClient.m_hWnd) {
		POINT pt = { GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) };

		ChatCtrl::sSelectedUser = Util::emptyStringT;

		int iBegin = 0, iEnd = 0;
		if(ctrlClient.HitNick(pt, ChatCtrl::sSelectedUser, iBegin, iEnd) == false) {
			return 0;
		}

		bHandled = true;
		ctrlClient.SetSel(iBegin, iEnd);
		ctrlClient.InvalidateRect(NULL);

		UserInfo* ui = findUser(ChatCtrl::sSelectedUser);
		if(ui != NULL) {
			if (wParam & MK_CONTROL) { // MK_CONTROL = 0x0008
				PrivateFrame::openWindow(ui->user);
			} else if (wParam & MK_SHIFT) {
				try {
					QueueManager::getInstance()->addList(ui->user, QueueItem::FLAG_CLIENT_VIEW);
				} catch(const Exception& e) {
					addClientLine(Text::toT(e.getError()), WinUtil::m_ChatTextSystem, true);
				}
			} else {
				switch(SETTING(CHAT_DBLCLICK)) {
					case 0: {
                        int items = ctrlUsers.GetItemCount();
                        int pos = -1;
					    ctrlUsers.SetRedraw(FALSE);
					    for(int i = 0; i < items; ++i) {
                            if(ctrlUsers.getItemData(i) == ui)
                                pos = i;
                            ctrlUsers.SetItemState(i, (i == pos) ? LVIS_SELECTED | LVIS_FOCUSED : 0, LVIS_SELECTED | LVIS_FOCUSED);
					    }
					    ctrlUsers.SetRedraw(TRUE);
					    ctrlUsers.EnsureVisible(pos, FALSE);
					    break;
					}    
					case 1: {
                        tstring sUser = tstring(ui->getText(COLUMN_NICK, NULL)) + _T(": ");
                        ctrlMessage.AppendText(sUser.c_str());
					    ctrlMessage.SetFocus();
					    break;
					}
					case 2:
						if(ui->user != ClientManager::getInstance()->getMe())
                            ui->pm();
                        break;
					case 3:
					    ui->getList();
					    break;
					case 4:
					    ui->matchQueue();
					    break;
					case 5:
					    ui->grant();
					    break;
					case 6:
					    ui->addFav();
					    break;
				}
			}
			return 0;
		}
        
        OnlineUser* ou = client->findUser(Text::fromT(ChatCtrl::sSelectedUser));
		if(ou != NULL) {
			if (wParam & MK_CONTROL) { // MK_CONTROL = 0x0008
				PrivateFrame::openWindow(ou->getUser());
			} else if (wParam & MK_SHIFT) {
				try {
					QueueManager::getInstance()->addList(ou->getUser(), QueueItem::FLAG_CLIENT_VIEW);
				} catch(const Exception& e) {
					addClientLine(Text::toT(e.getError()), WinUtil::m_ChatTextSystem, true);
				}
			} else {
				switch(SETTING(CHAT_DBLCLICK)) { 
					case 1: {
                        tstring sUser = ChatCtrl::sSelectedUser + _T(": ");
                        ctrlMessage.AppendText(sUser.c_str());
					    ctrlMessage.SetFocus();
					    break;
					}
					case 2:
						if(ou->getUser() != ClientManager::getInstance()->getMe())
                            PrivateFrame::openWindow(ou->getUser());
                        break;
					case 3:
                    	try {
                    		QueueManager::getInstance()->addList(ou->getUser(), QueueItem::FLAG_CLIENT_VIEW);
                    	} catch(const Exception& e) {
                    		LogManager::getInstance()->message(e.getError());
                    	}
					    break;
					case 4:
                    	try {
                    		QueueManager::getInstance()->addList(ou->getUser(), QueueItem::FLAG_MATCH_QUEUE);
                    	} catch(const Exception& e) {
                    		LogManager::getInstance()->message(e.getError());
                    	}
					    break;
					case 5:
					    UploadManager::getInstance()->reserveSlot(ou->getUser());
					    break;
					case 6:
					    FavoriteManager::getInstance()->addFavoriteUser(ou->getUser());
					    break;
				}
			}
			return 0;
		}
	}
	return 0;
}

void HubFrame::addLine(const Identity& i, const tstring& aLine, CHARFORMAT2& cf, bool bUseEmo/* = true*/) {
	ctrlClient.AdjustTextSize();

	if(BOOLSETTING(LOG_MAIN_CHAT)) {
		StringMap params;
		params["message"] = Text::fromT(aLine);
		client->getHubIdentity().getParams(params, "hub", false);
		params["hubURL"] = client->getHubUrl();
		client->getMyIdentity().getParams(params, "my", true);
		LOG(LogManager::CHAT, params);
	}
	
	ctrlClient.AppendText(i, Text::toT(client->getMyNick()), timeStamps ? Text::toT("[" + Util::getShortTimeString() + "] ") : _T(""), 
		(BOOLSETTING(IP_IN_CHAT) && client->isOp()) ? Text::toT(i.getIp()): Util::emptyStringT, aLine, cf, bUseEmo);

	if (BOOLSETTING(BOLD_HUB)) {
		setDirty();
	}
	ctrlLastLines.Activate(FALSE);
	ctrlLastLines.Activate(TRUE);
}

LRESULT HubFrame::onTabContextMenu(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM lParam, BOOL& /*bHandled*/) {
	POINT pt = { GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) };        // location of mouse click 
	tabMenuShown = true;
	ipMenuShown = false;
	bool delseparator = false;
	prepareMenu(tabMenu, ::UserCommand::CONTEXT_HUB, client->getHubUrl(), client->isOp());
	if(!(tabMenu.GetMenuState(tabMenu.GetMenuItemCount()-1, MF_BYPOSITION) & MF_SEPARATOR)) {	
		tabMenu.AppendMenu(MF_SEPARATOR);
		delseparator = true;
	}
	tabMenu.AppendMenu(MF_STRING, IDC_CLOSE_WINDOW, CTSTRING(CLOSE));
	tabMenu.TrackPopupMenu(TPM_LEFTALIGN | TPM_BOTTOMALIGN | TPM_RIGHTBUTTON, pt.x, pt.y, m_hWnd);
	if(delseparator == true) {
		tabMenu.DeleteMenu(tabMenu.GetMenuItemCount()-1, MF_BYPOSITION);
	}
	tabMenu.DeleteMenu(tabMenu.GetMenuItemCount()-1, MF_BYPOSITION);
	cleanMenu(tabMenu);
	return TRUE;
}

LRESULT HubFrame::onCtlColor(UINT /*uMsg*/, WPARAM wParam, LPARAM /*lParam*/, BOOL& /*bHandled*/) {
	HDC hDC = (HDC)wParam;
	::SetBkColor(hDC, WinUtil::bgColor);
	::SetTextColor(hDC, WinUtil::textColor);
	return (LRESULT)WinUtil::bgBrush;
}
	
LRESULT HubFrame::onContextMenu(UINT /*uMsg*/, WPARAM wParam, LPARAM lParam, BOOL& /*bHandled*/) {
	CRect rc;            // client area of window 
	POINT pt = { GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) };        // location of mouse click
	tabMenuShown = false;
	ipMenuShown = false;
	CMenu Mnu;
	ChatCtrl::sSelectedUser = Util::emptyStringT;

	ctrlUsers.GetHeader().GetWindowRect(&rc);
		
	if(PtInRect(&rc, pt) && ShowUserList) {
		ctrlUsers.showMenu(pt);
		return TRUE;
	}

	if(reinterpret_cast<HWND>(wParam) == ctrlUsers && ShowUserList) { 
		if(ctrlUsers.GetSelectedCount() == 1) {
			if(pt.x == -1 && pt.y == -1) {
				WinUtil::getContextMenuPos(ctrlUsers, pt);
			}
			int i = -1;
			i = ctrlUsers.GetNextItem(i, LVNI_SELECTED);
			if(i >= 0) {
				ChatCtrl::sSelectedUser = Text::toT(((UserInfo*)ctrlUsers.getItemData(i))->getNick());
			}
		}

		if(PreparePopupMenu(&ctrlUsers, ChatCtrl::sSelectedUser, &Mnu)) {
			prepareMenu(Mnu, ::UserCommand::CONTEXT_CHAT, client->getHubUrl(), client->isOp());
			Mnu.TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, pt.x, pt.y, m_hWnd);
			cleanMenu(Mnu);
			return TRUE;
		}
	}
	
	if(reinterpret_cast<HWND>(wParam) == ctrlClient) { 
		if(pt.x == -1 && pt.y == -1) {
			CRect erc;
			ctrlClient.GetRect(&erc);
			pt.x = erc.Width() / 2;
			pt.y = erc.Height() / 2;
			ctrlClient.ClientToScreen(&pt);
		}
		POINT ptCl = pt;
		ctrlClient.ScreenToClient(&ptCl);
		ctrlClient.OnRButtonDown(ptCl);
		
		bool boHitURL = ctrlClient.HitURL();
		if(!boHitURL)
			sSelectedURL = _T("");

		if(PreparePopupMenu(&ctrlClient, ChatCtrl::sSelectedUser, &Mnu)) {
			if(ChatCtrl::sSelectedUser.empty()) {
				Mnu.TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, pt.x, pt.y, m_hWnd);
				if(ipMenuShown == true) {
                    cleanMenu(Mnu);
                }
				return TRUE;
			} else {
				prepareMenu(Mnu, ::UserCommand::CONTEXT_CHAT, client->getHubUrl(), client->isOp());
				if(!(Mnu.GetMenuState(Mnu.GetMenuItemCount()-1, MF_BYPOSITION) & MF_SEPARATOR)) {	
					Mnu.AppendMenu(MF_SEPARATOR);
				}
				Mnu.AppendMenu(MF_STRING, ID_EDIT_CLEAR_ALL, CTSTRING(CLEAR));
				Mnu.TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, pt.x, pt.y, m_hWnd);
				cleanMenu(Mnu);
				return TRUE;
			}
		}
	}
	return FALSE;
}

void HubFrame::runUserCommand(::UserCommand& uc) {
	if(!WinUtil::getUCParams(m_hWnd, uc, ucLineParams))
		return;

	StringMap ucParams = ucLineParams;

	client->getMyIdentity().getParams(ucParams, "my", true);
	client->getHubIdentity().getParams(ucParams, "hub", false);

	if(tabMenuShown == true) {
		client->escapeParams(ucParams);
		if(BOOLSETTING(USER_COMMANDS)) {
			addClientLine(_T("User command: ") + Text::toT(Util::formatParams(uc.getCommand(), ucParams, false)), 
				WinUtil::m_ChatTextSystem, true);
		}

		client->sendUserCmd(Util::formatParams(uc.getCommand(), ucParams, false));
	} if(ipMenuShown == true) {
		if(ChatCtrl::sSelectedIP.empty() == false) {
			StringMap tmp = ucParams;
			tmp["userI4"] = Text::fromT(ChatCtrl::sSelectedIP);
			client->escapeParams(tmp);

			if(BOOLSETTING(USER_COMMANDS)) {
				addClientLine(_T("User command: ") + Text::toT(Util::formatParams(uc.getCommand(), tmp, false)), 
					WinUtil::m_ChatTextSystem, true);
			}

            client->sendUserCmd(Util::formatParams(uc.getCommand(), tmp, false)); 
        }
	} else {
		if(ChatCtrl::sSelectedUser.empty() == false) {
			int sel = ctrlUsers.findItem(ChatCtrl::sSelectedUser);
			if(sel != -1) { 
				UserInfo* u = ctrlUsers.getItemData(sel);
				if(u->user->isOnline()) {
					StringMap tmp = ucParams;
					u->getIdentity().getParams(tmp, "user", true);
					client->escapeParams(tmp); 
					if(BOOLSETTING(USER_COMMANDS)) {
						addClientLine(_T("User command: ") + Text::toT(Util::formatParams(uc.getCommand(), tmp, false)), 
							WinUtil::m_ChatTextSystem, true);
					}

					client->sendUserCmd(Util::formatParams(uc.getCommand(), tmp, false));
				}
				return;
			}
			
			OnlineUser* ou = client->findUser(Text::fromT(ChatCtrl::sSelectedUser));				
    		if(ou != NULL) {
				if(ou->getUser()->isOnline()) {
					StringMap tmp = ucParams;
					ou->getIdentity().getParams(tmp, "user", true);
					client->escapeParams(tmp); 
					if(BOOLSETTING(USER_COMMANDS)) {
						addClientLine(_T("User command: ") + Text::toT(Util::formatParams(uc.getCommand(), tmp, false)), 
							WinUtil::m_ChatTextSystem, true);
					}

					client->sendUserCmd(Util::formatParams(uc.getCommand(), tmp, false));
				}
                return;
            }
		} else {
			int sel = -1;
			while((sel = ctrlUsers.GetNextItem(sel, LVNI_SELECTED)) != -1) {
				UserInfo* u = ctrlUsers.getItemData(sel);
				if(u->user->isOnline()) {
					StringMap tmp = ucParams;
					u->getIdentity().getParams(tmp, "user", true);
					client->escapeParams(tmp);
					if(BOOLSETTING(USER_COMMANDS)) {
						addClientLine(_T("User command: ") + Text::toT(Util::formatParams(uc.getCommand(), tmp, false)), 
							WinUtil::m_ChatTextSystem, true);
					}

					client->sendUserCmd(Util::formatParams(uc.getCommand(), tmp, false));
				}
			}
		}
	}
	return;
}

void HubFrame::onTab() {
	HWND focus = GetFocus();
	if((focus == ctrlMessage.m_hWnd) && !WinUtil::isShift()) {
		int n = ctrlMessage.GetWindowTextLength();
		AutoArray<TCHAR> buf(n+1);
		ctrlMessage.GetWindowText(buf, n+1);
		tstring text(buf, n);
		string::size_type textStart = text.find_last_of(_T(" \n\t"));

		if(complete.empty()) {
			if(textStart != string::npos) {
				complete = text.substr(textStart + 1);
			} else {
				complete = text;
			}
			if(complete.empty()) {
				// Still empty, no text entered...
				ctrlUsers.SetFocus();
				return;
			}
			int y = ctrlUsers.GetItemCount();

			for(int x = 0; x < y; ++x)
				ctrlUsers.SetItemState(x, 0, LVNI_FOCUSED | LVNI_SELECTED);
		}

		if(textStart == string::npos)
			textStart = 0;
		else
			textStart++;

		if(inTabComplete) {
			// Already pressed tab once. Output nick candidate list.
			tstring nicks;
			for(StringList::const_iterator i = client->tabCompleteNicks.begin(); i < client->tabCompleteNicks.end(); i+=2)
				nicks.append(Text::toT(*i + " "));

			addClientLine(nicks, WinUtil::m_ChatTextSystem, true);
			inTabComplete = false;
		} else {
			// First tab. Maximally extend proposed nick.
			tstring nick = Text::toT(client->scanNickPrefix(Text::fromT(complete)));
			if(client->tabCompleteNicks.empty())
                return;

			// Maybe it found a unique match. If userlist showing, highlight.
			if(ShowUserList && client->tabCompleteNicks.size() == 2) {
				int i = ctrlUsers.findItem(Text::toT(client->tabCompleteNicks[1]));
				ctrlUsers.SetItemState(i, LVNI_FOCUSED | LVNI_SELECTED, LVNI_FOCUSED | LVNI_SELECTED);
				ctrlUsers.EnsureVisible(i, FALSE);
			}

			ctrlMessage.SetSel(textStart, ctrlMessage.GetWindowTextLength(), TRUE);
			// no shift, use partial nick when appropriate
			if(GetAsyncKeyState(VK_SHIFT) & 0x8000) {
				ctrlMessage.ReplaceSel(nick.c_str());
			} else {
				ctrlMessage.ReplaceSel(Text::toT(client->stripNick(Text::fromT(nick))).c_str());
			}

			inTabComplete = true;
		}
	}

	if(focus == ctrlClient.m_hWnd) {
		ctrlMessage.SetFocus();
	} else if(focus == ctrlUsers.m_hWnd) {
		ctrlFilter.SetFocus();
	} else if(focus == ctrlFilter.m_hWnd) {
		ctrlFilterSel.SetFocus();
	} else if(focus == ctrlFilterSel.m_hWnd) {
		ctrlClient.SetFocus();
	}
}

LRESULT HubFrame::onFileReconnect(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(!client->isConnected() || !BOOLSETTING(CONFIRM_HUB_RECONNECT) || MessageBox(CTSTRING(CONFIRM_HUBRECONNECT), _T(APPNAME) _T(" ") _T(VERSIONSTRING), MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2) == IDYES) {
	    client->reconnect();
	}
	return 0;
}

LRESULT HubFrame::onChar(UINT uMsg, WPARAM wParam, LPARAM /*lParam*/, BOOL& bHandled) {
	if (GetFocus() != ctrlMessage.m_hWnd) { // don't handle these keys unless the user is entering a message
		if (uMsg != WM_KEYDOWN) {
			if(wParam == VK_TAB) {
				bHandled = TRUE;
				return 0;
			}
		}
		if(wParam == VK_TAB) {
			onTab();
			return 0;
		}
		bHandled = FALSE;
		return 0;
	}
	if(!complete.empty() && wParam != VK_TAB && uMsg == WM_KEYDOWN) {
		complete.clear();
		inTabComplete = false;
	}

	if (uMsg != WM_KEYDOWN) {
		switch(wParam) {
			case VK_RETURN:
				if( (GetKeyState(VK_CONTROL) & 0x8000) || (GetKeyState(VK_MENU) & 0x8000) ) {
					bHandled = FALSE;
				}
				break;
			case VK_TAB:
				bHandled = TRUE;
  				break;
  			default:
  				bHandled = FALSE;
				break;
			}
		return 0;
	}

	switch(wParam) {
		case VK_TAB:
			onTab();
			break;
		case VK_RETURN:
			if( (GetKeyState(VK_CONTROL) & 0x8000) || 
				(GetKeyState(VK_MENU) & 0x8000) ) {
					bHandled = FALSE;
				} else {
						onEnter();
					}
			break;
		case VK_UP:
			if ( (GetKeyState(VK_MENU) & 0x8000) ||	( ((GetKeyState(VK_CONTROL) & 0x8000) == 0) ^ (BOOLSETTING(SETTINGS_USE_CTRL_FOR_LINE_HISTORY) == true) ) ) {
				//scroll up in chat command history
				//currently beyond the last command?
				if (curCommandPosition > 0) {
					//check whether current command needs to be saved
					if (curCommandPosition == prevCommands.size()) {
						auto_ptr<TCHAR> messageContents(new TCHAR[ctrlMessage.GetWindowTextLength()+2]);
						ctrlMessage.GetWindowText(messageContents.get(), ctrlMessage.GetWindowTextLength()+1);
						currentCommand = tstring(messageContents.get());
					}

					//replace current chat buffer with current command
					ctrlMessage.SetWindowText(prevCommands[--curCommandPosition].c_str());
				}
				// move cursor to end of line
				ctrlMessage.SetSel(ctrlMessage.GetWindowTextLength(), ctrlMessage.GetWindowTextLength());
			} else {
				bHandled = FALSE;
			}

			break;
		case VK_DOWN:
			if ( (GetKeyState(VK_MENU) & 0x8000) ||	( ((GetKeyState(VK_CONTROL) & 0x8000) == 0) ^ (BOOLSETTING(SETTINGS_USE_CTRL_FOR_LINE_HISTORY) == true) ) ) {
				//scroll down in chat command history

				//currently beyond the last command?
				if (curCommandPosition + 1 < prevCommands.size()) {
					//replace current chat buffer with current command
					ctrlMessage.SetWindowText(prevCommands[++curCommandPosition].c_str());
				} else if (curCommandPosition + 1 == prevCommands.size()) {
					//revert to last saved, unfinished command

					ctrlMessage.SetWindowText(currentCommand.c_str());
					++curCommandPosition;
				}
				// move cursor to end of line
				ctrlMessage.SetSel(ctrlMessage.GetWindowTextLength(), ctrlMessage.GetWindowTextLength());
			} else {
				bHandled = FALSE;
			}

			break;
		case VK_PRIOR: // page up
			ctrlClient.SendMessage(WM_VSCROLL, SB_PAGEUP);

			break;
		case VK_NEXT: // page down
			ctrlClient.SendMessage(WM_VSCROLL, SB_PAGEDOWN);

			break;
		case VK_HOME:
			if (!prevCommands.empty() && (GetKeyState(VK_CONTROL) & 0x8000) || (GetKeyState(VK_MENU) & 0x8000)) {
				curCommandPosition = 0;
				
				auto_ptr<TCHAR> messageContents(new TCHAR[ctrlMessage.GetWindowTextLength()+2]);
				ctrlMessage.GetWindowText(messageContents.get(), ctrlMessage.GetWindowTextLength()+1);
				currentCommand = tstring(messageContents.get());

				ctrlMessage.SetWindowText(prevCommands[curCommandPosition].c_str());
			} else {
				bHandled = FALSE;
			}

			break;
		case VK_END:
			if ((GetKeyState(VK_CONTROL) & 0x8000) || (GetKeyState(VK_MENU) & 0x8000)) {
				curCommandPosition = prevCommands.size();

				ctrlMessage.SetWindowText(currentCommand.c_str());
			} else {
				bHandled = FALSE;
				}
				break;
		default:
			bHandled = FALSE;
	}
	return 0;
}

LRESULT HubFrame::onShowUsers(UINT /*uMsg*/, WPARAM wParam, LPARAM /*lParam*/, BOOL& bHandled) {
	bHandled = FALSE;
	if((wParam == BST_CHECKED)) {
		ShowUserList = true;
		client->GetUserList();
	} else {
		ShowUserList = false;
		ctrlUsers.DeleteAllItems();
        for(UserMapIter i = userMap.begin(); i != userMap.end(); ++i) {
            delete i->second;
            i->second = NULL;
        }
	}

	UpdateLayout(FALSE);
	return 0;
}

LRESULT HubFrame::onFollow(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(!redirect.empty()) {
		if(ClientManager::getInstance()->isConnected(Text::fromT(redirect))) {
			addClientLine(TSTRING(REDIRECT_ALREADY_CONNECTED), WinUtil::m_ChatTextServer, true);
			return 0;
		}
		
		dcassert(frames.find(server) != frames.end());
		dcassert(frames[server] == this);
		frames.erase(server);
		server = redirect;
		frames[server] = this;

		// the client is dead, long live the client!
		client->removeListener(this);
		ClientManager::getInstance()->putClient(client);
		clearUserList();
		clearTaskList();
		client = ClientManager::getInstance()->getClient(Text::fromT(server));
		ctrlClient.SetClient(client);

		RecentHubEntry r;
		r.setName("*");
		r.setDescription("***");
		r.setUsers("*");
		r.setShared("*");
		r.setServer(Text::fromT(redirect));
		FavoriteManager::getInstance()->addRecent(r);

		client->addListener(this);
		client->connect();
	}
	return 0;
}

LRESULT HubFrame::onEnterUsers(int /*idCtrl*/, LPNMHDR /* pnmh */, BOOL& /*bHandled*/) {
	int item = ctrlUsers.GetNextItem(-1, LVNI_FOCUSED);
	if(item != -1) {
		try {
			QueueManager::getInstance()->addList((ctrlUsers.getItemData(item))->user, QueueItem::FLAG_CLIENT_VIEW);
		} catch(const Exception& e) {
			addClientLine(Text::toT(e.getError()), WinUtil::m_ChatTextSystem, true);
		}
	}
	return 0;
}

LRESULT HubFrame::onGetToolTip(int /*idCtrl*/, LPNMHDR pnmh, BOOL& /*bHandled*/) {
	NMTTDISPINFO* nm = (NMTTDISPINFO*)pnmh;
	lastLines.clear();
	for(TStringIter i = lastLinesList.begin(); i != lastLinesList.end(); ++i) {
		lastLines += *i;
		lastLines += _T("\r\n");
	}
	if(lastLines.size() > 2) {
		lastLines.erase(lastLines.size() - 2);
	}
	nm->lpszText = const_cast<TCHAR*>(lastLines.c_str());
	return 0;
}


void HubFrame::resortUsers() {
	for(FrameIter i = frames.begin(); i != frames.end(); ++i)
		i->second->resortForFavsFirst(true);
}

void HubFrame::closeDisconnected() {
	for(FrameIter i=frames.begin(); i!= frames.end(); ++i) {
		if (!(i->second->client->isConnected())) {
			i->second->PostMessage(WM_CLOSE);
		}
	}
}

void HubFrame::on(FavoriteManagerListener::UserAdded, const FavoriteUser& /*aUser*/) throw() {
	resortForFavsFirst();
}
void HubFrame::on(FavoriteManagerListener::UserRemoved, const FavoriteUser& /*aUser*/) throw() {
	resortForFavsFirst();
}

void HubFrame::resortForFavsFirst(bool justDoIt /* = false */) {
	if(justDoIt || BOOLSETTING(SORT_FAVUSERS_FIRST)) {
		speak(RESORT_USERLIST);
	}
}

void HubFrame::on(Second, uint64_t /*aTick*/) throw() {   
	if(updateUsers) {
		updateUsers = false;
		if(m_hWnd) {
            speak(STATS);
        } else {
            PostMessage(WM_SPEAKER);
        }
	} else if(m_hWnd) {
        speak(STATS);
    }
}

void HubFrame::on(Connecting, Client*) throw() { 
	if(BOOLSETTING(SEARCH_PASSIVE) && (SETTING(CONNECTION_TYPE) == SettingsManager::CONNECTION_ACTIVE)) {
		addSystemLine(TSTRING(ANTI_PASSIVE_SEARCH), WinUtil::m_ChatTextSystem);
	}
	speak(ADD_STATUS_LINE, STRING(CONNECTING_TO) + client->getHubUrl() + "...");
	speak(SET_WINDOW_TITLE, client->getHubUrl());
}
void HubFrame::on(Connected, Client*) throw() { 
	speak(CONNECTED);
}
void HubFrame::on(UserUpdated, Client*, const OnlineUser& user) throw() { 
	speak(UPDATE_USER_JOIN, user);
}
void HubFrame::on(UsersUpdated, Client*, const OnlineUser::List& aList) throw() {
	for(OnlineUser::List::const_iterator i = aList.begin(); i != aList.end(); ++i) {
		tasks.add(UPDATE_USER, new UserTask(*(*i)));
	}
	updateUsers = true;
}

void HubFrame::on(ClientListener::UserRemoved, Client*, const OnlineUser& user) throw() {
    speak(REMOVE_USER, user);
}

void HubFrame::on(Redirect, Client*, const string& line) throw() { 
	if(ClientManager::getInstance()->isConnected(line)) {
		speak(ADD_STATUS_LINE, STRING(REDIRECT_ALREADY_CONNECTED));
		return;
	}

	redirect = Text::toT(line);
	if(BOOLSETTING(AUTO_FOLLOW)) {
		PostMessage(WM_COMMAND, IDC_FOLLOW, 0);
	} else {
		speak(ADD_STATUS_LINE, STRING(PRESS_FOLLOW) + line);
	}
}
void HubFrame::on(Failed, Client*, const string& line) throw() { 
	speak(ADD_STATUS_LINE, line); 
	speak(DISCONNECTED); 
}
void HubFrame::on(GetPassword, Client*) throw() { 
	speak(GET_PASSWORD);
}
void HubFrame::on(HubUpdated, Client*) throw() {
	string hubName = client->getHubName();
	if(!client->getHubDescription().empty()) {
		hubName += " - " + client->getHubDescription();
	}
	hubName += " (" + client->getHubUrl() + ")";
	speak(SET_WINDOW_TITLE, hubName);
}
void HubFrame::on(Message, Client*, const OnlineUser& u, const string& line) throw() {
    char *chatline = (char *)line.c_str();
    if(&u == NULL || u.getIdentity().isOp()) {
        char *temp;
        if((temp = strchr(chatline, '\n')) != NULL) {
            temp[0] = NULL;
            if((strstr(chatline, "is kicking") != NULL) && (strstr(chatline, "because:") != NULL) ||
    		  (strstr(chatline, "was kicked by") != NULL)) {
                temp[0] = '\n';
				speak(KICK_MSG, u, Util::toDOS(line));
                return;
            }
            temp[0] = '\n';
    	} else if((strstr(chatline, "is kicking") != NULL) && (strstr(chatline, "because:") != NULL) ||
    		(strstr(chatline, "was kicked by") != NULL)) {
            speak(KICK_MSG, u, Util::toDOS(line));
            return;
    	}
    }

	speak(ADD_CHAT_LINE, u, Util::toDOS(line));
}

void HubFrame::on(PrivateMessage, Client*, const OnlineUser& user, const OnlineUser& by, const string& line) throw() { 
	speak(PRIVATE_MESSAGE, user, by, line);
}

void HubFrame::on(PrivateMessageNick, Client*, const string& nick, const OnlineUser& by, const string& line) throw() {
	speak(PRIVATE_MESSAGE_NICK, nick, by, line);
}

void HubFrame::on(NickTaken, Client*) throw() { 
	speak(ADD_STATUS_LINE, STRING(NICK_TAKEN));
}
void HubFrame::on(SearchFlood, Client*, const string& line) throw() {
	speak(ADD_STATUS_LINE, STRING(SEARCH_SPAM_FROM) + line);
}

void HubFrame::addClientLine(const tstring& aLine, const CHARFORMAT2& cf, const bool &inChat, bool bNoBold/* = false*/) {
	tstring line = _T("[") + Text::toT(Util::getShortTimeString()) + _T("] ") + aLine;
    TCHAR* sLine = (TCHAR*)line.c_str();
    
    if(_tcslen(sLine) > 512) {
        sLine[512] = NULL;
    }

	ctrlStatus.SetText(0, sLine);
	while(lastLinesList.size() + 1 > MAX_CLIENT_LINES)
		lastLinesList.erase(lastLinesList.begin());
		
	lastLinesList.push_back(sLine);
	
	if(BOOLSETTING(BOLD_HUB) == true && bNoBold == false) {
		setDirty();
	}
	
	if(BOOLSETTING(STATUS_IN_CHAT) && inChat) {
		ctrlClient.AppendStatusText(timeStamps ? Text::toT("[" + Util::getShortTimeString() + "] ") : _T(""), _T("*** ") + aLine, (CHARFORMAT2&)cf);
	}

	if(BOOLSETTING(LOG_STATUS_MESSAGES)) {
		StringMap params;
		client->getHubIdentity().getParams(params, "hub", false);
		params["hubURL"] = client->getHubUrl();
		client->getMyIdentity().getParams(params, "my", true);
		params["message"] = Text::fromT(aLine);
		LOG(LogManager::STATUS, params);
	}
}

void HubFrame::addSystemLine(const tstring& aLine, const CHARFORMAT2& cf) {
    ctrlClient.AppendStatusText(timeStamps ? Text::toT("[" + Util::getShortTimeString() + "] ") : _T(""), aLine, (CHARFORMAT2&)cf);
}

LRESULT HubFrame::onMouseMove(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM lParam, BOOL& /*bHandled*/) {
	RECT rc;
	POINT pt = { GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) };        // location of mouse click
	// Get the bounding rectangle of the client area. 
	ctrlClient.GetClientRect(&rc);
	ctrlClient.ScreenToClient(&pt); 
	if (PtInRect(&rc, pt)) { 
		sSelectedURL = _T("");
  		return TRUE;
	}
	return 1;
}

bool HubFrame::PreparePopupMenu(CWindow *pCtrl, tstring& sNick, CMenu *pMenu) {
	if(pMenu->m_hMenu != NULL) {
		pMenu->DestroyMenu();
		pMenu->m_hMenu = NULL;
	}

	pMenu->CreatePopupMenu();

	bool bIsChat = (pCtrl == ((CWindow*)&ctrlClient));
	if(sNick.empty() && bIsChat == true) {
		if(ChatCtrl::sSelectedIP.empty() == false) {
			ipMenuShown = true;
			pMenu->AppendMenu(MF_STRING, IDC_WHOIS_IP, (CTSTRING(WHO_IS) + ChatCtrl::sSelectedIP).c_str() );
			pMenu->SetMenuDefaultItem(IDC_WHOIS_IP);
			
			prepareMenu(*pMenu, ::UserCommand::CONTEXT_IP, client->getHubUrl(), client->isOp());
			if((pMenu->GetMenuState(pMenu->GetMenuItemCount()-1, MF_BYPOSITION) & MF_SEPARATOR) == false) {	
				pMenu->AppendMenu(MF_SEPARATOR);
			}
		}
		long lSelBegin, lSelEnd; ctrlClient.GetSel(lSelBegin, lSelEnd);
        if(lSelEnd > lSelBegin) {
            if(searchMenu.m_hMenu != NULL) {
        		searchMenu.DestroyMenu();
        		searchMenu.m_hMenu = NULL;
        	}
        	searchMenu.CreatePopupMenu();
        	searchMenu.AppendMenu(MF_STRING, IDC_SEARCH_DC, CTSTRING(SEARCH_DC));
        	searchMenu.AppendMenu(MF_STRING, IDC_SEARCH_GOOGLE, CTSTRING(SEARCH_GOOGLE));
        	searchMenu.AppendMenu(MF_STRING, IDC_SEARCH_YAHOO, CTSTRING(SEARCH_YAHOO));
        	searchMenu.AppendMenu(MF_STRING, IDC_SEARCH_IMDB, CTSTRING(SEARCH_IMDB));
        	searchMenu.AppendMenu(MF_STRING, IDC_SEARCH_CSFD, CTSTRING(SEARCH_CSFD));
        	searchMenu.AppendMenu(MF_STRING, IDC_SEARCH_WIKI, CTSTRING(SEARCH_WIKI));
        	pMenu->AppendMenu(MF_POPUP, (UINT)(HMENU)searchMenu, CTSTRING(SEARCH));
            pMenu->AppendMenu(MF_SEPARATOR);
        }
		pMenu->AppendMenu(MF_STRING, ID_EDIT_COPY, CTSTRING(COPY));
		pMenu->AppendMenu(MF_STRING, IDC_COPY_ACTUAL_LINE,  CTSTRING(COPY_LINE));
		if(sSelectedURL.empty() == false) 
  			pMenu->AppendMenu(MF_STRING, IDC_COPY_URL, CTSTRING(COPY_URL));
		pMenu->AppendMenu(MF_SEPARATOR);
		pMenu->AppendMenu(MF_STRING, ID_EDIT_SELECT_ALL, CTSTRING(SELECT_ALL));
		pMenu->AppendMenu(MF_STRING, ID_EDIT_CLEAR_ALL, CTSTRING(CLEAR));
		pMenu->AppendMenu(MF_SEPARATOR);
		pMenu->AppendMenu(MF_STRING, IDC_AUTOSCROLL_CHAT, CTSTRING(ASCROLL_CHAT));

		if(ctrlClient.GetAutoScroll())
			pMenu->CheckMenuItem(IDC_AUTOSCROLL_CHAT, MF_BYCOMMAND | MF_CHECKED);
	} else {
		if(sNick.empty() == false) {
		    bool bIsMe = (sNick == Text::toT(client->getMyNick()));
			// Jediny nick
			if(bIsChat) {
				tstring sTmp = TSTRING(USER) + _T(" ") + sNick;
				if(!BOOLSETTING(LOG_PRIVATE_CHAT)) {
					pMenu->AppendMenu(MF_STRING | MF_DISABLED, (UINT_PTR)0, sTmp.c_str());
				} else {
					pMenu->AppendMenu(MF_STRING, IDC_OPEN_USER_LOG, sTmp.c_str());
				}
				pMenu->AppendMenu(MF_SEPARATOR);
			}
			if(bIsChat == false && bIsMe == false && BOOLSETTING(LOG_PRIVATE_CHAT) == true) {
				pMenu->AppendMenu(MF_STRING, IDC_OPEN_USER_LOG, CTSTRING(OPEN_USER_LOG));
				pMenu->AppendMenu(MF_SEPARATOR);
			}
			if(bIsMe == false) {
				pMenu->AppendMenu(MF_STRING, IDC_PUBLIC_MESSAGE, CTSTRING(SEND_PUBLIC_MESSAGE));
				pMenu->AppendMenu(MF_STRING, IDC_PRIVATEMESSAGE, CTSTRING(SEND_PRIVATE_MESSAGE));
			}
			if(bIsChat == true && ShowUserList == true) {
				pMenu->AppendMenu(MF_STRING, IDC_SELECT_USER, CTSTRING(SELECT_USER_LIST));
			}
			if(bIsMe == false) {
			   pMenu->AppendMenu(MF_SEPARATOR);
			}
            
            if(copyMenu.m_hMenu != NULL) {
        		copyMenu.DestroyMenu();
        		copyMenu.m_hMenu = NULL;
        	}
        	copyMenu.CreatePopupMenu();
        	copyMenu.AppendMenu(MF_STRING, IDC_COPY_NICK, CTSTRING(COPY_NICK));
        	copyMenu.AppendMenu(MF_STRING, IDC_COPY_EXACT_SHARE, CTSTRING(COPY_EXACT_SHARE));
        	copyMenu.AppendMenu(MF_STRING, IDC_COPY_DESCRIPTION, CTSTRING(COPY_DESCRIPTION));
        	copyMenu.AppendMenu(MF_STRING, IDC_COPY_TAG, CTSTRING(COPY_TAG));
        	copyMenu.AppendMenu(MF_STRING, IDC_COPY_EMAIL_ADDRESS, CTSTRING(COPY_EMAIL_ADDRESS));
        	if(client->isOp()) {
        		copyMenu.AppendMenu(MF_STRING, IDC_COPY_IP, CTSTRING(COPY_IP));
        	}
    	   
			pMenu->AppendMenu(MF_POPUP, (UINT)(HMENU)copyMenu, CTSTRING(COPY));
			if(bIsMe == false) {
                if(grantMenu.m_hMenu != NULL) {
            		grantMenu.DestroyMenu();
            		grantMenu.m_hMenu = NULL;
            	}
            	grantMenu.CreatePopupMenu();
            	grantMenu.AppendMenu(MF_STRING, IDC_GRANTSLOT, CTSTRING(GRANT_EXTRA_SLOT));
            	grantMenu.AppendMenu(MF_STRING, IDC_GRANTSLOT_HOUR, CTSTRING(GRANT_EXTRA_SLOT_HOUR));
            	grantMenu.AppendMenu(MF_STRING, IDC_GRANTSLOT_DAY, CTSTRING(GRANT_EXTRA_SLOT_DAY));
            	grantMenu.AppendMenu(MF_STRING, IDC_GRANTSLOT_WEEK, CTSTRING(GRANT_EXTRA_SLOT_WEEK));
            	grantMenu.AppendMenu(MF_SEPARATOR);
            	grantMenu.AppendMenu(MF_STRING, IDC_UNGRANTSLOT, CTSTRING(REMOVE_EXTRA_SLOT));
	
				pMenu->AppendMenu(MF_POPUP, (UINT)(HMENU)grantMenu, CTSTRING(GRANT_SLOTS_MENU));
				int i = ctrlUsers.findItem(sNick);
				if(i != -1) {
					UserInfo* ui = (UserInfo*)ctrlUsers.getItemData(i);
					if(client->isOp() || !ui->getIdentity().isOp()) {
						pMenu->AppendMenu(MF_SEPARATOR);
						if(!ignoreList.count(sNick)) {
						      pMenu->AppendMenu(MF_STRING, IDC_IGNORE, CTSTRING(IGNORE_USER));
						} else {    
						      pMenu->AppendMenu(MF_STRING, IDC_UNIGNORE, CTSTRING(UNIGNORE_USER));
						}    
					}
				} else {
                    OnlineUser* ou = client->findUser(Text::fromT(sNick));				
    				if(ou != NULL) {
    					if(client->isOp() || !ou->getIdentity().isOp()) {
    						pMenu->AppendMenu(MF_SEPARATOR);
    						if(!ignoreList.count(sNick)) {
    						      pMenu->AppendMenu(MF_STRING, IDC_IGNORE, CTSTRING(IGNORE_USER));
    						} else {    
    						      pMenu->AppendMenu(MF_STRING, IDC_UNIGNORE, CTSTRING(UNIGNORE_USER));
    						}    
    					}
                    }
                }
				pMenu->AppendMenu(MF_SEPARATOR);
				pMenu->AppendMenu(MF_STRING, IDC_GETLIST, CTSTRING(GET_FILE_LIST));
				pMenu->AppendMenu(MF_STRING, IDC_MATCH_QUEUE, CTSTRING(MATCH_QUEUE));
				pMenu->AppendMenu(MF_STRING, IDC_ADD_TO_FAVORITES, CTSTRING(ADD_TO_FAVORITES));
			}
		} else {
			// Muze byt vice nicku
			if(bIsChat == false) {
				// Pocet oznacenych
				int iCount = ctrlUsers.GetSelectedCount();
				TCHAR sTmp[256];
				_stprintf_s(sTmp, 256, _T("%i user(s)"), iCount);
				pMenu->AppendMenu(MF_STRING | MF_DISABLED, (UINT_PTR)0, sTmp);
			}
			if (ctrlUsers.GetSelectedCount() <= 25) {
				pMenu->AppendMenu(MF_SEPARATOR);
				pMenu->AppendMenu(MF_STRING, IDC_PUBLIC_MESSAGE, CTSTRING(SEND_PUBLIC_MESSAGE));
			}
			pMenu->AppendMenu(MF_SEPARATOR);
			pMenu->AppendMenu(MF_STRING, IDC_GETLIST, CTSTRING(GET_FILE_LIST));
			pMenu->AppendMenu(MF_STRING, IDC_MATCH_QUEUE, CTSTRING(MATCH_QUEUE));
			pMenu->AppendMenu(MF_STRING, IDC_ADD_TO_FAVORITES, CTSTRING(ADD_TO_FAVORITES));
		}
		if(bIsChat) {
			switch(SETTING(CHAT_DBLCLICK)) {
		          case 0:
						pMenu->SetMenuDefaultItem(IDC_SELECT_USER);
						break;
		          case 1:
						pMenu->SetMenuDefaultItem(IDC_PUBLIC_MESSAGE);
						break;
		          case 2:
						pMenu->SetMenuDefaultItem(IDC_PRIVATEMESSAGE);
						break;
		          case 3:
						pMenu->SetMenuDefaultItem(IDC_GETLIST);
						break;
		          case 4:
						pMenu->SetMenuDefaultItem(IDC_MATCH_QUEUE);
						break;
		          case 6:
						pMenu->SetMenuDefaultItem(IDC_ADD_TO_FAVORITES);
						break;
			}    
		} else {    
			switch(SETTING(USERLIST_DBLCLICK)) {
		          case 0:
						pMenu->SetMenuDefaultItem(IDC_GETLIST);
						break;
		          case 1:
						pMenu->SetMenuDefaultItem(IDC_PUBLIC_MESSAGE);
						break;
		          case 2:
						pMenu->SetMenuDefaultItem(IDC_PRIVATEMESSAGE);
						break;
		          case 3:
						pMenu->SetMenuDefaultItem(IDC_MATCH_QUEUE);
						break;
		          case 5:
						pMenu->SetMenuDefaultItem(IDC_ADD_TO_FAVORITES);
						break;
			}
		}   		
	}
	return true;
}

LRESULT HubFrame::onCopyActualLine(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if((GetFocus() == ctrlClient.m_hWnd) && (ChatCtrl::sSelectedLine != _T(""))) {
		WinUtil::setClipboard(ChatCtrl::sSelectedLine);
	}
	return 0;
}

LRESULT HubFrame::onSelectUser(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(ChatCtrl::sSelectedUser.empty()) {
		return 0;
	}

	int pos = ctrlUsers.findItem(ChatCtrl::sSelectedUser);
	if(pos == -1) {
		return 0;
	}

	int items = ctrlUsers.GetItemCount();
	ctrlUsers.SetRedraw(FALSE);
	for(int i = 0; i < items; ++i) {
		ctrlUsers.SetItemState(i, (i == pos) ? LVIS_SELECTED | LVIS_FOCUSED : 0, LVIS_SELECTED | LVIS_FOCUSED);
	}
	ctrlUsers.SetRedraw(TRUE);
	ctrlUsers.EnsureVisible(pos, FALSE);
	return 0;
}

LRESULT HubFrame::onPublicMessage(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(!client->isConnected())
		return 0;
		
	tstring sUsers = Util::emptyStringT;

	if(ChatCtrl::sSelectedUser.empty() == false) {
		sUsers = ChatCtrl::sSelectedUser;
	} else {
        int i = -1;
		while((i = ctrlUsers.GetNextItem(i, LVNI_SELECTED)) != -1) {
			if(sUsers.size() > 0)
  				sUsers += _T(", ");
  				
			sUsers += Text::toT(((UserInfo*)ctrlUsers.getItemData(i))->getNick());
		}
	}

	sUsers += _T(": ");
	ctrlMessage.AppendText(sUsers.c_str());
	ctrlMessage.SetFocus();
	return 0;
}

LRESULT HubFrame::onAutoScrollChat(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	ctrlClient.SetAutoScroll( !ctrlClient.GetAutoScroll() );
	return 0;
}

LRESULT HubFrame::onCopyURL(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(sSelectedURL != _T("")) {
		WinUtil::setClipboard(sSelectedURL);
	}
	return 0;
}

LRESULT HubFrame::onClientEnLink(int /*idCtrl*/, LPNMHDR pnmh, BOOL& /*bHandled*/) {
	ENLINK* pEL = (ENLINK*)pnmh;

	if(pEL->msg == WM_LBUTTONUP) {
		long lBegin = pEL->chrg.cpMin, lEnd = pEL->chrg.cpMax;

		TCHAR* sURLTemp = new TCHAR[(lEnd - lBegin)+1];
		if(sURLTemp == NULL) {
			return 0;
		}
    	ctrlClient.GetTextRange(lBegin, lEnd, sURLTemp);
    	tstring sURL = sURLTemp;
    	WinUtil::openLink(sURL);
		delete[] sURLTemp;
	} else if(pEL->msg == WM_RBUTTONUP) {
		sSelectedURL = _T("");
		long lBegin = pEL->chrg.cpMin, lEnd = pEL->chrg.cpMax;

		TCHAR* sURLTemp = new TCHAR[(lEnd - lBegin)+1];
		if(sURLTemp == NULL) {
			return 0;
		}
    	ctrlClient.GetTextRange(lBegin, lEnd, sURLTemp);
    	sSelectedURL = sURLTemp;
		delete[] sURLTemp;

		ctrlClient.SetSel(lBegin, lEnd);
		ctrlClient.InvalidateRect(NULL);
		return 0;
	}
	return 0;
}

LRESULT HubFrame::onOpenUserLog(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {	
	StringMap params;
	UserInfo* ui = NULL;
	if(ChatCtrl::sSelectedUser.empty() == false) {
		int k = ctrlUsers.findItem(ChatCtrl::sSelectedUser);
		if(k != -1)
            ui = ctrlUsers.getItemData(k);
	} else {
		int i = -1;
		if((i = ctrlUsers.GetNextItem(i, LVNI_SELECTED)) != -1) {
			ui = (UserInfo*)ctrlUsers.getItemData(i);
		}
	}

	if(ui != NULL) {
        params["userNI"] = ui->getNick();
        params["userCID"] = ui->user->getCID().toBase32();
    } else if(ChatCtrl::sSelectedUser.empty() == false) {
        OnlineUser* ou = client->findUser(Text::fromT(ChatCtrl::sSelectedUser));				
    	if(ou != NULL) {
            params["userNI"] = ou->getIdentity().getNick();
            params["userCID"] = ou->getUser()->getCID().toBase32();
        }
    }

	params["hubNI"] = client->getHubName();
	params["myNI"] = client->getMyNick();
	params["hubURL"] = client->getHubUrl();
	tstring file = Text::toT(Util::validateFileName(SETTING(LOG_DIRECTORY) + Util::formatParams(SETTING(LOG_FILE_PRIVATE_CHAT), params, true)));
	if(File::existsFile(Text::fromT(file))) {
		ShellExecute(NULL, NULL, file.c_str(), NULL, NULL, SW_SHOWNORMAL);
	} else {
		MessageBox(CTSTRING(NO_LOG_FOR_USER),CTSTRING(NO_LOG_FOR_USER), MB_OK );	  
	}
	return 0;
}

LRESULT HubFrame::onOpenHubLog(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	StringMap params;
	params["hubNI"] = client->getHubName();
	params["hubURL"] = client->getHubUrl();
	params["myNI"] = client->getMyNick(); 
	tstring filename = Text::toT(Util::validateFileName(SETTING(LOG_DIRECTORY) + Util::formatParams(SETTING(LOG_FILE_MAIN_CHAT), params, true)));
	if(File::existsFile(Text::fromT(filename))){
		ShellExecute(NULL, NULL, filename.c_str(), NULL, NULL, SW_SHOWNORMAL);
	} else {
		MessageBox(CTSTRING(NO_LOG_FOR_HUB),CTSTRING(NO_LOG_FOR_HUB), MB_OK );	  
	}
	return 0;
}

LRESULT HubFrame::onWhoisIP(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(ChatCtrl::sSelectedIP != _T("")) {
 		WinUtil::openLink(_T("http://www.ripe.net/perl/whois?form_type=simple&full_query_string=&searchtext=") + ChatCtrl::sSelectedIP);
 	}
	return 0;
}

LRESULT HubFrame::onStyleChange(UINT /*uMsg*/, WPARAM wParam, LPARAM /*lParam*/, BOOL& bHandled) {
	bHandled = FALSE;
	if((wParam & MK_LBUTTON) && ::GetCapture() == m_hWnd) {
		UpdateLayout(FALSE);
	}
	return 0;
}

LRESULT HubFrame::onStyleChanged(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& bHandled) {
	bHandled = FALSE;
	UpdateLayout(FALSE);
	return 0;
}

LRESULT HubFrame::onFilterChar(UINT /*uMsg*/, WPARAM wParam, LPARAM /*lParam*/, BOOL& bHandled) {
	switch(wParam) {
		case VK_RETURN: {
            AutoArray<TCHAR> buf(ctrlFilter.GetWindowTextLength()+1);
			ctrlFilter.GetWindowText(buf, ctrlFilter.GetWindowTextLength());
			filterstring = Text::fromT(tstring(buf, ctrlFilter.GetWindowTextLength()));
			filtercolumn = ctrlFilterSel.GetCurSel();
			updateUserList();
			bHandled = FALSE;
			break; }
		case VK_TAB:
			onTab();
			break;
		default:
			bHandled = FALSE;
			break;
	}
	return 0;
}

void HubFrame::updateUserList() {
    ctrlUsers.SetRedraw(FALSE);
  	
    ctrlUsers.DeleteAllItems();

    if(filterstring.empty()) {
    	for(UserMap::iterator i = userMap.begin(); i != userMap.end(); ++i){
    		if(i->second != NULL) {
    			i->second->update(i->second->getIdentity(), ctrlUsers.getSortColumn());
    			ctrlUsers.insertItem(i->second, WinUtil::getImage(i->second->getIdentity()));
    		}
    	}
    } else {
        for(UserMap::iterator i = userMap.begin(); i != userMap.end(); ++i){
        	if(i->second != NULL) {
        		if(filterUser(i->second->getIdentity()) == true) {
        			ctrlUsers.insertItem(i->second, WinUtil::getImage(i->second->getIdentity()));
        		}
        	}
        }
    }
	ctrlUsers.SetRedraw(TRUE);
} 

bool HubFrame::filterUser(const Identity& id) {
	switch(filtercolumn) {
		case 0:
			//Nick
			if(Util::stristr(id.getNick().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 1:
			//Shared
			if(Util::stristr(Util::formatBytes(id.getBytesShared()).c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 2:
			//Exact Shared
			if(Util::stristr(Util::toString(id.getBytesShared()).c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 3:
			//Description
			if(Util::stristr(Text::toUtf8(id.getDescription()).c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 4:
			//Tag
			if(Util::stristr(Text::toUtf8(id.getTag()).c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 5:
			//Connection
			if(Util::stristr(Text::toUtf8(id.getConnection()).c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 6:
			//Email
			if(Util::stristr(Text::toUtf8(id.getEmail()).c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 7:
			//Client ID
			if(Util::stristr(Text::toUtf8(id.getClientID()).c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 8:
			//Version
			if(Util::stristr(Text::toUtf8(id.getVersion()).c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 9:
			//Mode
			if(Util::stristr(id.getMode().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 10:
			//Hubs
			if(Util::stristr(id.getHubs().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 11:
			//Slots
			if(Util::stristr(Text::toUtf8(id.getSlots()).c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 12:
			//Limit
			if(Util::stristr(id.getUpload().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 13:
			//IP
			if(Util::stristr(id.getIp().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		default:
			return false;
			break;
	}
	return false;
}

void HubFrame::on(SettingsManagerListener::Save, SimpleXML& /*xml*/) throw() {
    PostMessage(WM_REFRESH_SETTINGS);
}

LRESULT HubFrame::onRefreshSettings(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/) {
	if(ctrlUsers.GetBkColor() != WinUtil::bgColor) {
		ctrlClient.SetBackgroundColor(WinUtil::bgColor);
		ctrlUsers.SetBkColor(WinUtil::bgColor);
		ctrlUsers.SetTextBkColor(WinUtil::bgColor);
	}
	if(ctrlUsers.GetTextColor() != WinUtil::textColor) {
		ctrlUsers.SetTextColor(WinUtil::textColor);
	}
	RedrawWindow(NULL, NULL, RDW_ERASE | RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN);   
	return 0;
}

LRESULT HubFrame::onCopyHubName(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
    WinUtil::setClipboard(Text::toT(client->getHubName()));
	return 0;
}
    
LRESULT HubFrame::onCopyHubAddress(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
    WinUtil::setClipboard(Text::toT(client->getHubUrl()));
	return 0;    
}    

LRESULT HubFrame::onCustomDraw(int /*idCtrl*/, LPNMHDR pnmh, BOOL& bHandled) {
	LPNMLVCUSTOMDRAW cd = (LPNMLVCUSTOMDRAW)pnmh;

	if(!BOOLSETTING(USERLIST_COLORING)) {
		bHandled = FALSE;
		return 0;
	}

	switch(cd->nmcd.dwDrawStage) {
		case CDDS_PREPAINT:
			return CDRF_NOTIFYITEMDRAW;

		case CDDS_ITEMPREPAINT: {
		    UserInfo* ui = (UserInfo*)cd->nmcd.lItemlParam; 
			if(FavoriteManager::getInstance()->isFavoriteUser(ui->user)){
				cd->clrText = SETTING(FAVORITE_COLOR);
			} else if(UploadManager::getInstance()->hasReservedSlot(ui->user)) {
				cd->clrText = SETTING(EXTRA_SLOT_COLOR);
			} else if(ignoreList.count(Text::toT(ui->getIdentity().getNick()))) {
				cd->clrText = SETTING(IGNORED_COLOR);
			} else if(ui->getIdentity().isSet(Identity::FIREBALL)) {
				cd->clrText = SETTING(FIREBALL_COLOR);
			} else if(ui->getIdentity().isSet(Identity::SERVER)) {
				cd->clrText = SETTING(SERVER_COLOR);
			} else if(ui->getIdentity().isOp()) {
				cd->clrText = SETTING(OP_COLOR);
			} else if(ui->getIdentity().getMode() == "A" && !ui->user->isSet(User::PASSIVE)) {
				cd->clrText = SETTING(ACTIVE_COLOR);
			} else if(ui->getIdentity().getUser()->isSet(User::PASSIVE) || ui->getIdentity().getMode() == "P" || ui->getIdentity().getMode() == "5") {
				cd->clrText = SETTING(PASIVE_COLOR);
			} else {
				cd->clrText = SETTING(NORMAL_COLOR);
			}
			return CDRF_NEWFONT | CDRF_NOTIFYSUBITEMDRAW;
		}

	default:
		return CDRF_DODEFAULT;
	}
}

LRESULT HubFrame::onSearch(WORD /*wNotifyCode*/, WORD wID, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
    long lSelBegin, lSelEnd; ctrlClient.GetSel(lSelBegin, lSelEnd);
    if(lSelEnd > lSelBegin) { 
        long iLen = lSelEnd - lSelBegin; 
        AutoArray<TCHAR> sTemp(iLen+1);
        ctrlClient.GetTextRange(lSelBegin, lSelEnd, sTemp);
        sTemp[iLen] = 0;
        tstring sSearchString = tstring(sTemp, iLen);
    	switch (wID) {
    		case IDC_SEARCH_DC:
    			WinUtil::SearchOnDC(sSearchString);
    			break;
    		case IDC_SEARCH_GOOGLE:
    			WinUtil::SearchOnGoogle(sSearchString);
    			break;
    		case IDC_SEARCH_YAHOO:
    			WinUtil::SearchOnYahoo(sSearchString);
    			break;
    		case IDC_SEARCH_IMDB:
    			WinUtil::SearchOnIMDB(sSearchString);
    			break;
    		case IDC_SEARCH_CSFD:
    			WinUtil::SearchOnCSFD(sSearchString);
    			break;
    		default:
    			break;
    	}
    }
	return 0;
}

LRESULT HubFrame::onIgnore(UINT /*uMsg*/, WPARAM /*wParam*/, HWND /*lParam*/, BOOL& /*bHandled*/) {
	if(client->isConnected()) {
		if(ChatCtrl::sSelectedUser != Util::emptyStringT) {
			ignoreList.insert(ChatCtrl::sSelectedUser);
		} else {
            int i = -1;
			while((i = ctrlUsers.GetNextItem(i, LVNI_SELECTED)) != -1) {
				ignoreList.insert(Text::toT(((UserInfo*)ctrlUsers.getItemData(i))->getNick()));
			}
		}
	}
	return 0;
}

LRESULT HubFrame::onUnignore(UINT /*uMsg*/, WPARAM /*wParam*/, HWND /*lParam*/, BOOL& /*bHandled*/) {
	if(client->isConnected()) {
		if(ChatCtrl::sSelectedUser != Util::emptyStringT) {
			ignoreList.erase(ChatCtrl::sSelectedUser);
		} else {
            int i = -1;
			while((i = ctrlUsers.GetNextItem(i, LVNI_SELECTED)) != -1) {
				ignoreList.erase(Text::toT(((UserInfo*)ctrlUsers.getItemData(i))->getNick()));
			}
		}
	}
	return 0;
}
	
LRESULT HubFrame::onMatchQueue(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(ChatCtrl::sSelectedUser != Util::emptyStringT) {
		int i = ctrlUsers.findItem(ChatCtrl::sSelectedUser);
		if(i != -1) {
			ctrlUsers.forEachAtPos(i, &UserInfoBase::matchQueue);
		} else {
            OnlineUser* ou = client->findUser(Text::fromT(ChatCtrl::sSelectedUser));				
    		if(ou != NULL) {
                try {
                	QueueManager::getInstance()->addList(ou->getUser(), QueueItem::FLAG_MATCH_QUEUE);
                } catch(const Exception& e) {
                	LogManager::getInstance()->message(e.getError());
                }
            }
        }
	} else {
		ctrlUsers.forEachSelected(&UserInfoBase::matchQueue);
	}
	return 0;
}

LRESULT HubFrame::onGetList(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(ChatCtrl::sSelectedUser != Util::emptyStringT) {
		int i = ctrlUsers.findItem(ChatCtrl::sSelectedUser);
		if(i != -1) {
			ctrlUsers.forEachAtPos(i, &UserInfoBase::getList);
		} else {
            OnlineUser* ou = client->findUser(Text::fromT(ChatCtrl::sSelectedUser));				
    		if(ou != NULL) {
                try {
                	QueueManager::getInstance()->addList(ou->getUser(), QueueItem::FLAG_CLIENT_VIEW);
                } catch(const Exception& e) {
                	LogManager::getInstance()->message(e.getError());
                }
            }
        }
	} else {
		ctrlUsers.forEachSelected(&UserInfoBase::getList);
	}
	return 0;
}

LRESULT HubFrame::onAddToFavorites(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(ChatCtrl::sSelectedUser != Util::emptyStringT) {
		int i = ctrlUsers.findItem(ChatCtrl::sSelectedUser);
		if(i != -1) {
			ctrlUsers.forEachAtPos(i, &UserInfoBase::addFav);
		} else {
            OnlineUser* ou = client->findUser(Text::fromT(ChatCtrl::sSelectedUser));				
    		if(ou != NULL) {
                FavoriteManager::getInstance()->addFavoriteUser(ou->getUser());
            }
        }
	} else {
		ctrlUsers.forEachSelected(&UserInfoBase::addFav);
	}
	return 0;
}

LRESULT HubFrame::onPrivateMessage(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(ChatCtrl::sSelectedUser != Util::emptyStringT) {
		int i = ctrlUsers.findItem(ChatCtrl::sSelectedUser);
		if(i != -1) {
			ctrlUsers.forEachAtPos(i, &UserInfoBase::pm);
		} else {
            OnlineUser* ou = client->findUser(Text::fromT(ChatCtrl::sSelectedUser));				
    		if(ou != NULL) {
                PrivateFrame::openWindow(ou->getUser());
            }
        }
	} else {
		ctrlUsers.forEachSelected(&UserInfoBase::pm);
	}
	return 0;
}

LRESULT HubFrame::onGrantSlot(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) { 
	if(ChatCtrl::sSelectedUser != Util::emptyStringT) {
		int i = ctrlUsers.findItem(ChatCtrl::sSelectedUser);
		if(i != -1) {
			ctrlUsers.forEachAtPos(i, &UserInfoBase::grant);
		} else {
            OnlineUser* ou = client->findUser(Text::fromT(ChatCtrl::sSelectedUser));				
    		if(ou != NULL) {
                UploadManager::getInstance()->reserveSlot(ou->getUser());
            }
        }
	} else {
		ctrlUsers.forEachSelected(&UserInfoBase::grant);
	}
	return 0;
}

LRESULT HubFrame::onGrantSlotHour(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) { 
	if(ChatCtrl::sSelectedUser != Util::emptyStringT) {
		int i = ctrlUsers.findItem(ChatCtrl::sSelectedUser);
		if(i != -1) {
			ctrlUsers.forEachAtPos(i, &UserInfoBase::grantSlotHour);
		} else {
            OnlineUser* ou = client->findUser(Text::fromT(ChatCtrl::sSelectedUser));				
    		if(ou != NULL) {
                UploadManager::getInstance()->reserveSlotHour(ou->getUser());
            }
        }
	} else {
		ctrlUsers.forEachSelected(&UserInfoBase::grantSlotHour);
	}
	return 0;
}

LRESULT HubFrame::onGrantSlotDay(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) { 
	if(ChatCtrl::sSelectedUser != Util::emptyStringT) {
		int i = ctrlUsers.findItem(ChatCtrl::sSelectedUser);
		if(i != -1) {
			ctrlUsers.forEachAtPos(i, &UserInfoBase::grantSlotDay);
		} else {
            OnlineUser* ou = client->findUser(Text::fromT(ChatCtrl::sSelectedUser));				
    		if(ou != NULL) {
                UploadManager::getInstance()->reserveSlotDay(ou->getUser());
            }
        }
	} else {
		ctrlUsers.forEachSelected(&UserInfoBase::grantSlotDay);
	}
	return 0;
}

LRESULT HubFrame::onGrantSlotWeek(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) { 
	if(ChatCtrl::sSelectedUser != Util::emptyStringT) {
		int i = ctrlUsers.findItem(ChatCtrl::sSelectedUser);
		if(i != -1) {
			ctrlUsers.forEachAtPos(i, &UserInfoBase::grantSlotWeek);
		} else {
            OnlineUser* ou = client->findUser(Text::fromT(ChatCtrl::sSelectedUser));				
    		if(ou != NULL) {
                UploadManager::getInstance()->reserveSlotWeek(ou->getUser());
            }
        }
	} else {
		ctrlUsers.forEachSelected(&UserInfoBase::grantSlotWeek);
	}
	return 0;
}

LRESULT HubFrame::onUnGrantSlot(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) { 
	if(ChatCtrl::sSelectedUser != Util::emptyStringT) {
		int i = ctrlUsers.findItem(ChatCtrl::sSelectedUser);
		if(i != -1) {
			ctrlUsers.forEachAtPos(i, &UserInfoBase::ungrantSlot );
		} else {
            OnlineUser* ou = client->findUser(Text::fromT(ChatCtrl::sSelectedUser));				
    		if(ou != NULL) {
                UploadManager::getInstance()->unreserveSlot(ou->getUser());
            }
        }
	} else {
		ctrlUsers.forEachSelected(&UserInfoBase::ungrantSlot);
	}
	return 0;
}
	
/* Some parts Copyright (C) 2002-2008 Petr Kozelka, PPK at czdc dot org */
