/*
 * Copyright (C) 2001-2007 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include "../client/DCPlusPlus.h"
#include "Resource.h"

#include "../client/ResourceManager.h"
#include "../client/QueueManager.h"
#include "../client/ConnectionManager.h"
#include "../client/ClientManager.h"

#include "WinUtil.h"
#include "TransferView.h"
#include "MainFrm.h"

int TransferView::columnIndexes[] = { COLUMN_USER, COLUMN_HUB, COLUMN_STATUS, COLUMN_TIMELEFT, COLUMN_SPEED, COLUMN_FILE, COLUMN_SIZE, COLUMN_PATH, COLUMN_IP, COLUMN_RATIO };
int TransferView::columnSizes[] = { 150, 100, 250, 75, 75, 175, 100, 200, 50, 75 };

static ResourceManager::Strings columnNames[] = { ResourceManager::USER, ResourceManager::HUB, ResourceManager::STATUS,
ResourceManager::TIME_LEFT, ResourceManager::SPEED, ResourceManager::FILENAME, ResourceManager::SIZE, ResourceManager::PATH,
ResourceManager::COUNTRY, ResourceManager::RATIO};

TransferView::~TransferView() {
	arrows.Destroy();
}

LRESULT TransferView::onCreate(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/) {
	if(Util::isXP) {
        arrows.CreateFromImage(IDB_ARROWS_XP, 16, 16, CLR_DEFAULT, IMAGE_BITMAP, LR_CREATEDIBSECTION | LR_SHARED);
	} else {
        arrows.CreateFromImage(IDB_ARROWS, 16, 16, CLR_DEFAULT, IMAGE_BITMAP, LR_CREATEDIBSECTION | LR_SHARED);
	}


	ctrlTransfers.Create(m_hWnd, rcDefault, NULL, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | 
		WS_HSCROLL | WS_VSCROLL | LVS_REPORT | LVS_SHOWSELALWAYS | LVS_SHAREIMAGELISTS, WS_EX_CLIENTEDGE, IDC_TRANSFERS);
	
	if(Util::isXP) {
		ctrlTransfers.SetExtendedListViewStyle(LVS_EX_HEADERDRAGDROP | LVS_EX_FULLROWSELECT);
	} else {
		ctrlTransfers.SetExtendedListViewStyle(LVS_EX_HEADERDRAGDROP | LVS_EX_FULLROWSELECT | 0x00010000);
	}

	WinUtil::splitTokens(columnIndexes, SETTING(MAINFRAME_ORDER), COLUMN_LAST);
	WinUtil::splitTokens(columnSizes, SETTING(MAINFRAME_WIDTHS), COLUMN_LAST);

	for(int j = 0; j < COLUMN_LAST; j++) {
		int fmt = (j == COLUMN_SIZE || j == COLUMN_TIMELEFT || j == COLUMN_SPEED) ? LVCFMT_RIGHT : LVCFMT_LEFT;
		ctrlTransfers.InsertColumn(j, CTSTRING_I(columnNames[j]), fmt, columnSizes[j], j);
	}

	ctrlTransfers.setColumnOrderArray(COLUMN_LAST, columnIndexes);
	ctrlTransfers.setVisible(SETTING(MAINFRAME_VISIBLE));

	ctrlTransfers.SetBkColor(WinUtil::bgColor);
	ctrlTransfers.SetTextBkColor(WinUtil::bgColor);
	ctrlTransfers.SetTextColor(WinUtil::textColor);
	ctrlTransfers.setOwnerDraw(true);

	ctrlTransfers.SetImageList(arrows, LVSIL_SMALL);
	ctrlTransfers.setSortColumn(COLUMN_USER);

	ConnectionManager::getInstance()->addListener(this);
	DownloadManager::getInstance()->addListener(this);
	UploadManager::getInstance()->addListener(this);
	SettingsManager::getInstance()->addListener(this);
	
	return 0;
}

void TransferView::prepareClose() {
	ctrlTransfers.saveHeaderOrder(SettingsManager::MAINFRAME_ORDER, SettingsManager::MAINFRAME_WIDTHS,
		SettingsManager::MAINFRAME_VISIBLE);
	
	ConnectionManager::getInstance()->removeListener(this);
	DownloadManager::getInstance()->removeListener(this);
	UploadManager::getInstance()->removeListener(this);
	SettingsManager::getInstance()->removeListener(this);
}

LRESULT TransferView::onSize(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/) {
	RECT rc;
	GetClientRect(&rc);
	ctrlTransfers.MoveWindow(&rc);

	return 0;
}

LRESULT TransferView::onContextMenu(UINT /*uMsg*/, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	if(reinterpret_cast<HWND>(wParam) == ctrlTransfers) {
		RECT rc;                    // client area of window 
		POINT pt = { GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) };

		ctrlTransfers.GetHeader().GetWindowRect(&rc);
		if(PtInRect(&rc, pt)){
			ctrlTransfers.showMenu(pt);
			return TRUE;
		}

		if(pt.x == -1 && pt.y == -1) {
			WinUtil::getContextMenuPos(ctrlTransfers, pt);
		}
		
		if(ctrlTransfers.GetSelectedCount() == 1) { 
			if(transferMenu.m_hMenu != NULL) {
				transferMenu.DestroyMenu();
				transferMenu.m_hMenu = NULL;
			}
			if(grantMenu.m_hMenu != NULL) {
				grantMenu.DestroyMenu();
				grantMenu.m_hMenu = NULL;
			}
			if(previewMenu.m_hMenu != NULL) {
				previewMenu.DestroyMenu();
				previewMenu.m_hMenu = NULL;
			}
	
			grantMenu.CreatePopupMenu();
			grantMenu.AppendMenu(MF_STRING, IDC_GRANTSLOT, CTSTRING(GRANT_EXTRA_SLOT));
			grantMenu.AppendMenu(MF_STRING, IDC_GRANTSLOT_HOUR, CTSTRING(GRANT_EXTRA_SLOT_HOUR));
			grantMenu.AppendMenu(MF_STRING, IDC_GRANTSLOT_DAY, CTSTRING(GRANT_EXTRA_SLOT_DAY));
			grantMenu.AppendMenu(MF_STRING, IDC_GRANTSLOT_WEEK, CTSTRING(GRANT_EXTRA_SLOT_WEEK));
			grantMenu.AppendMenu(MF_SEPARATOR);
			grantMenu.AppendMenu(MF_STRING, IDC_UNGRANTSLOT, CTSTRING(REMOVE_EXTRA_SLOT));
	
			int i = -1;
			ItemInfo* ii;
			transferMenu.CreatePopupMenu();
			if((i = ctrlTransfers.GetNextItem(i, LVNI_SELECTED)) != -1) {
				ii = ctrlTransfers.getItemData(i);
				if(ii->download == true) {
					transferMenu.AppendMenu(MF_STRING, IDC_SEARCH_ALTERNATES, CTSTRING(SEARCH_FOR_ALTERNATES));

					previewMenu.CreatePopupMenu();
					string ext = Util::getFileExt(ii->Target);
					if(ext.size()>1) ext = ext.substr(1);
					PreviewAppsSize = WinUtil::SetupPreviewMenu(previewMenu, ext);
					if(previewMenu.GetMenuItemCount() > 0) {
						if(previewMenu.GetMenuItemCount() == 1) // ugly but easy way to don't have submenu for only one item ;o)
							PreviewAppsSize = WinUtil::SetupPreviewMenu(transferMenu, ext);
						else
							transferMenu.AppendMenu(MF_POPUP, (UINT)(HMENU)previewMenu, CTSTRING(PREVIEW_MENU));
							
						transferMenu.AppendMenu(MF_SEPARATOR);
					}
				}
				if(BOOLSETTING(LOG_PRIVATE_CHAT)) {
					transferMenu.AppendMenu(MF_STRING, IDC_OPEN_USER_LOG, CTSTRING(OPEN_USER_LOG));
				}
				transferMenu.AppendMenu(MF_STRING, IDC_GETLIST, CTSTRING(GET_FILE_LIST));
				transferMenu.AppendMenu(MF_STRING, IDC_MATCH_QUEUE, CTSTRING(MATCH_QUEUE));
				transferMenu.AppendMenu(MF_STRING, IDC_PRIVATEMESSAGE, CTSTRING(SEND_PRIVATE_MESSAGE));
				transferMenu.AppendMenu(MF_STRING, IDC_ADD_TO_FAVORITES, CTSTRING(ADD_TO_FAVORITES));
				transferMenu.AppendMenu(MF_POPUP, (UINT)(HMENU)grantMenu, CTSTRING(GRANT_SLOTS_MENU));
				transferMenu.AppendMenu(MF_STRING, IDC_REMOVEALL, CTSTRING(REMOVE_FROM_ALL));
				transferMenu.AppendMenu(MF_STRING, IDC_COPY_NICK, CTSTRING(COPY_NICK));
				OnlineUser& ou = ClientManager::getInstance()->getOnlineUser(ii->user);
				if(&ou && ou.getClient().isOp()) {
					if(ii->IP != _T("")) {
						transferMenu.AppendMenu(MF_STRING, IDC_COPY_IP, CTSTRING(COPY_IP));
						transferMenu.AppendMenu(MF_STRING, IDC_WHOIS_IP, (CTSTRING(WHO_IS) + ii->IP).c_str());
					}
				}
				switch(SETTING(TRANSFERLIST_DBLCLICK)) {
					case 0:
			          transferMenu.SetMenuDefaultItem(IDC_PRIVATEMESSAGE);
			          break;
					case 1:
			          transferMenu.SetMenuDefaultItem(IDC_GETLIST);
			          break;
					case 2:
			          transferMenu.SetMenuDefaultItem(IDC_MATCH_QUEUE);
			          break;
					case 4:
			          transferMenu.SetMenuDefaultItem(IDC_ADD_TO_FAVORITES);
			          break;
				}
				StringList userhubs = ClientManager::getInstance()->getHubs(ii->user->getCID());
				if(userhubs.size() > 0) {
					prepareMenu(transferMenu, UserCommand::CONTEXT_CHAT, userhubs[0], ClientManager::getInstance()->isOp(ClientManager::getInstance()->getMe(), userhubs[0]));
				} else {
					prepareMenu(transferMenu, UserCommand::CONTEXT_CHAT, "", false);
				}
				transferMenu.AppendMenu(MF_SEPARATOR);
				if(ii->download == true) {
					transferMenu.AppendMenu(MF_STRING, IDC_FORCE, CTSTRING(FORCE_ATTEMPT));
					transferMenu.AppendMenu(MF_SEPARATOR);
				}
				transferMenu.AppendMenu(MF_STRING, IDC_REMOVE, CTSTRING(CLOSE_CONNECTION));
			}
			transferMenu.TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, pt.x, pt.y, m_hWnd);

			cleanMenu(transferMenu);
			
			return TRUE;
		} else if(ctrlTransfers.GetSelectedCount() > 0) {
			if(transferMenu.m_hMenu != NULL) {
				transferMenu.DestroyMenu();
				transferMenu.m_hMenu = NULL;
			}
		
			transferMenu.CreatePopupMenu();
			transferMenu.AppendMenu(MF_STRING, IDC_GETLIST, CTSTRING(GET_FILE_LIST));
			transferMenu.AppendMenu(MF_STRING, IDC_MATCH_QUEUE, CTSTRING(MATCH_QUEUE));
			transferMenu.AppendMenu(MF_STRING, IDC_ADD_TO_FAVORITES, CTSTRING(ADD_TO_FAVORITES));
			transferMenu.AppendMenu(MF_STRING, IDC_REMOVEALL, CTSTRING(REMOVE_FROM_ALL));
			switch(SETTING(TRANSFERLIST_DBLCLICK)) {
				case 0:
			        break;
				case 1:
			        transferMenu.SetMenuDefaultItem(IDC_GETLIST);
			        break;
				case 2:
			        transferMenu.SetMenuDefaultItem(IDC_MATCH_QUEUE);
			        break;
				case 4:
			        transferMenu.SetMenuDefaultItem(IDC_ADD_TO_FAVORITES);
			        break;
			}
			string hubaddress;
			bool bIsOP = true, bOneHub = true;
			int i = -1;
			while((i = ctrlTransfers.GetNextItem(i, LVNI_SELECTED)) != -1) {
				ItemInfo* ii = ctrlTransfers.getItemData(i);
				StringList userhubs = ClientManager::getInstance()->getHubs(ii->user->getCID());
            	string userhub;
            	if(userhubs.size() > 0) userhub = userhubs[0];

            	if(bOneHub && hubaddress.empty()) {
            		hubaddress = userhub;
            	} else if(hubaddress != userhub) {
            		bOneHub = false;
            		hubaddress.clear();
            	}
            	if(bIsOP)
					bIsOP = userhub.empty() ? false : ClientManager::getInstance()->isOp(ClientManager::getInstance()->getMe(), userhub);
            }
            prepareMenu(transferMenu, UserCommand::CONTEXT_CHAT, hubaddress, bIsOP);
			transferMenu.AppendMenu(MF_SEPARATOR);
			transferMenu.AppendMenu(MF_STRING, IDC_REMOVE, CTSTRING(CLOSE_CONNECTION));
			transferMenu.TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, pt.x, pt.y, m_hWnd);

			cleanMenu(transferMenu);
			
			return TRUE;
        }
	}
	bHandled = FALSE;
	return FALSE; 
}

void TransferView::runUserCommand(UserCommand& uc) {
	if(!WinUtil::getUCParams(m_hWnd, uc, ucLineParams))
		return;

	StringMap ucParams = ucLineParams;

	int i = -1;
	while((i = ctrlTransfers.GetNextItem(i, LVNI_SELECTED)) != -1) {
		ItemInfo* itemI = ctrlTransfers.getItemData(i);
		if(!itemI->user->isOnline())
			continue;

		StringMap tmp = ucParams;
		ucParams["fileFN"] = itemI->Target;

		// compatibility with 0.674 and earlier
		ucParams["file"] = ucParams["fileFN"];

		ClientManager::getInstance()->userCommand(itemI->user, uc, tmp, true);
	}
}

LRESULT TransferView::onForce(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	int i = -1;
	while( (i = ctrlTransfers.GetNextItem(i, LVNI_SELECTED)) != -1) {
		ItemInfo *ii = ctrlTransfers.getItemData(i);
		ii->columns[COLUMN_STATUS] = CTSTRING(CONNECTING_FORCED);
		ctrlTransfers.updateItem(ii);
		ConnectionManager::getInstance()->force(ii->user);
	}
	return 0;
}

void TransferView::ItemInfo::removeAll() {
	QueueManager::getInstance()->removeSource(user, QueueItem::Source::FLAG_REMOVED);
}

LRESULT TransferView::onCopyNick(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	int i = -1;
	while( (i = ctrlTransfers.GetNextItem(i, LVNI_SELECTED)) != -1) {
		WinUtil::setClipboard(WinUtil::getNicks(ctrlTransfers.getItemData(i)->user));
	}
	return 0;
}

LRESULT TransferView::onCustomDraw(int /*idCtrl*/, LPNMHDR pnmh, BOOL& bHandled) {
	CRect rc;
	NMLVCUSTOMDRAW* cd = (NMLVCUSTOMDRAW*)pnmh;

	switch(cd->nmcd.dwDrawStage) {
	case CDDS_PREPAINT:
		return CDRF_NOTIFYITEMDRAW;

	case CDDS_ITEMPREPAINT:
		return CDRF_NOTIFYSUBITEMDRAW;

	case CDDS_SUBITEM | CDDS_ITEMPREPAINT: {
		ItemInfo* ii = reinterpret_cast<ItemInfo*>(cd->nmcd.lItemlParam);

		int colIndex = ctrlTransfers.findColumn(cd->iSubItem);
		cd->clrTextBk = WinUtil::bgColor;

		if((ii->status == ItemInfo::STATUS_RUNNING) && (colIndex == COLUMN_STATUS) ) {
			if(!BOOLSETTING(SHOW_PROGRESS_BARS)) {
				bHandled = FALSE;
				return 0;
			}

			// draw something nice...
			COLORREF barBase = ii->download ? SETTING(DOWNLOAD_BAR_COLOR) : SETTING(UPLOAD_BAR_COLOR);
			COLORREF bgBase = WinUtil::bgColor;
			int mod = (HLS_L(RGB2HLS(bgBase)) >= 128) ? -30 : 30;
			COLORREF barPal[3] = { HLS_TRANSFORM(barBase, -40, 50), barBase, HLS_TRANSFORM(barBase, 40, -30) };
			COLORREF bgPal[2] = { HLS_TRANSFORM(bgBase, mod, 0), HLS_TRANSFORM(bgBase, mod/2, 0) };

			//this is just severely broken, msdn says GetSubItemRect requires a one based index
			//but it wont work and index 0 gives the rect of the whole item
			if(cd->iSubItem == 0) {
				//use LVIR_LABEL to exclude the icon area since we will be painting over that
				//later
				ctrlTransfers.GetItemRect((int)cd->nmcd.dwItemSpec, &rc, LVIR_LABEL);
			} else {
				ctrlTransfers.GetSubItemRect((int)cd->nmcd.dwItemSpec, cd->iSubItem, LVIR_BOUNDS, &rc);
			}
				
			CRect rc2 = rc;
			rc2.left += 6;
				
			// draw background
			HGDIOBJ oldpen = ::SelectObject(cd->nmcd.hdc, CreatePen(PS_SOLID,0,bgPal[0]));
			HGDIOBJ oldbr = ::SelectObject(cd->nmcd.hdc, CreateSolidBrush(bgPal[1]));
			::Rectangle(cd->nmcd.hdc, rc.left, rc.top - 1, rc.right, rc.bottom);			
			rc.DeflateRect(1, 0, 1, 1);

			LONG left = rc.left;
			int64_t w = rc.Width();
			// draw start part
			if(ii->size == 0)
				ii->size = 1;
			rc.right = left + (int) (w * ii->start / ii->size);
			DeleteObject(SelectObject(cd->nmcd.hdc, CreateSolidBrush(barPal[0])));
			DeleteObject(SelectObject(cd->nmcd.hdc, CreatePen(PS_SOLID,0,barPal[0])));
				
			::Rectangle(cd->nmcd.hdc, rc.left, rc.top, rc.right, rc.bottom);
				
			// Draw actual part
			rc.left = rc.right;
			rc.right = left + (int) (w * ii->actual / ii->size);
			DeleteObject(SelectObject(cd->nmcd.hdc, CreateSolidBrush(barPal[1])));

			::Rectangle(cd->nmcd.hdc, rc.left, rc.top, rc.right, rc.bottom);

			// And the effective part...
			if(ii->pos > ii->actual) {
				rc.left = rc.right - 1;
				rc.right = left + (int) (w * ii->pos / ii->size);
				DeleteObject(SelectObject(cd->nmcd.hdc, CreateSolidBrush(barPal[2])));

				::Rectangle(cd->nmcd.hdc, rc.left, rc.top, rc.right, rc.bottom);

			}
			rc.left = left;
			// draw progressbar highlight
			if(rc.Width()>2) {
				DeleteObject(SelectObject(cd->nmcd.hdc, CreatePen(PS_SOLID,1,barPal[2])));

				rc.top += 2;
				::MoveToEx(cd->nmcd.hdc,rc.left+1,rc.top,(LPPOINT)NULL);
				::LineTo(cd->nmcd.hdc,rc.right-2,rc.top);
			}
								
			// draw status text
			DeleteObject(::SelectObject(cd->nmcd.hdc, oldpen));
			DeleteObject(::SelectObject(cd->nmcd.hdc, oldbr));

			LONG right = rc2.right;
			left = rc2.left;
			rc2.right = rc.right;
			LONG top = rc2.top + (rc2.Height() - WinUtil::getTextHeight(cd->nmcd.hdc) - 1)/2;
			int textcolor = ii->download ? SETTING(PROGRESS_TEXT_COLOR_DOWN) : SETTING(PROGRESS_TEXT_COLOR_UP);
			SetBkMode(cd->nmcd.hdc, TRANSPARENT);
			SetTextColor(cd->nmcd.hdc, textcolor);
			::ExtTextOut(cd->nmcd.hdc, left, top, ETO_CLIPPED, rc2, ii->columns[COLUMN_STATUS].c_str(),
				ii->columns[COLUMN_STATUS].size(), NULL);

			rc2.left = rc2.right;
			rc2.right = right;

			SetTextColor(cd->nmcd.hdc, WinUtil::textColor);
			::ExtTextOut(cd->nmcd.hdc, left, top, ETO_CLIPPED, rc2, ii->columns[COLUMN_STATUS].c_str(),
				ii->columns[COLUMN_STATUS].size(), NULL);

			//bah crap, if we return CDRF_SKIPDEFAULT windows won't paint the icons
			//so we have to do it
			if(cd->iSubItem == 0){
				LVITEM lvItem;
				lvItem.iItem = cd->nmcd.dwItemSpec;
				lvItem.iSubItem = 0;
				lvItem.mask = LVIF_IMAGE | LVIF_STATE;
				lvItem.stateMask = LVIS_SELECTED;
				ctrlTransfers.GetItem(&lvItem);

				HIMAGELIST imageList = (HIMAGELIST)::SendMessage(ctrlTransfers.m_hWnd, LVM_GETIMAGELIST, LVSIL_SMALL, 0);
				if(imageList) {
					//let's find out where to paint it
					//and draw the background to avoid having 
					//the selection color as background
					CRect iconRect;
					ctrlTransfers.GetSubItemRect((int)cd->nmcd.dwItemSpec, 0, LVIR_ICON, iconRect);

					//we don't need to paint the background if the item's not selected
					//since it will already have the right color
					if(lvItem.state & LVIS_SELECTED) {
						HBRUSH brush = (HBRUSH)::SendMessage(::GetParent(ctrlTransfers.m_hWnd), WM_CTLCOLORLISTBOX, (WPARAM)cd->nmcd.hdc, (LPARAM)ctrlTransfers.m_hWnd);
						if(brush) {
							//remove 4 pixels to repaint the offset between the
							//column border and the icon.
							iconRect.left -= 4;
							::FillRect(cd->nmcd.hdc, &iconRect, brush);

							//have to add them back otherwise the icon will be painted
							//in the wrong place
							iconRect.left += 4;
						}
					}
					ImageList_Draw(imageList, lvItem.iImage, cd->nmcd.hdc, iconRect.left, iconRect.top, ILD_TRANSPARENT);
				}
			}
			return CDRF_SKIPDEFAULT;
		} else if(BOOLSETTING(GET_USER_COUNTRY) && (colIndex == COLUMN_IP)) {
			ItemInfo* ii = (ItemInfo*)cd->nmcd.lItemlParam;
			ctrlTransfers.GetSubItemRect((int)cd->nmcd.dwItemSpec, cd->iSubItem, LVIR_BOUNDS, rc);
			COLORREF color;
			if(ctrlTransfers.GetItemState((int)cd->nmcd.dwItemSpec, LVIS_SELECTED) & LVIS_SELECTED) {
				if(ctrlTransfers.m_hWnd == ::GetFocus()) {
					color = GetSysColor(COLOR_HIGHLIGHT);
					SetBkColor(cd->nmcd.hdc, GetSysColor(COLOR_HIGHLIGHT));
					SetTextColor(cd->nmcd.hdc, GetSysColor(COLOR_HIGHLIGHTTEXT));
				} else {
					color = GetBkColor(cd->nmcd.hdc);
					SetBkColor(cd->nmcd.hdc, color);
				}				
			} else {
				color = WinUtil::bgColor;
				SetBkColor(cd->nmcd.hdc, WinUtil::bgColor);
				SetTextColor(cd->nmcd.hdc, cd->clrText);
			}
			CRect rc2 = rc;
			rc2.left += 2;
			HGDIOBJ oldpen = ::SelectObject(cd->nmcd.hdc, CreatePen(PS_SOLID,0, color));
			HGDIOBJ oldbr = ::SelectObject(cd->nmcd.hdc, CreateSolidBrush(color));
			Rectangle(cd->nmcd.hdc, rc.left, rc.top, rc.right, rc.bottom);

			DeleteObject(::SelectObject(cd->nmcd.hdc, oldpen));
			DeleteObject(::SelectObject(cd->nmcd.hdc, oldbr));

			TCHAR buf[256];
			ctrlTransfers.GetItemText((int)cd->nmcd.dwItemSpec, cd->iSubItem, buf, 255);
			buf[255] = 0;
			if(_tcslen(buf) > 0) {
				LONG top = rc2.top + (rc2.Height() - 15)/2;
				if((top - rc2.top) < 2)
					top = rc2.top + 1;

				POINT p = { rc2.left, top };
				WinUtil::flagImages.Draw(cd->nmcd.hdc, ii->flagImage, p, LVSIL_SMALL);
				top = rc2.top + (rc2.Height() - WinUtil::getTextHeight(cd->nmcd.hdc) - 1)/2;
				::ExtTextOut(cd->nmcd.hdc, rc2.left + 29, top + 1, ETO_CLIPPED, rc2, buf, _tcslen(buf), NULL);
				return CDRF_SKIPDEFAULT;
			}
		} else if(((colIndex == COLUMN_FILE) || (colIndex == COLUMN_SIZE) || (colIndex == COLUMN_PATH)) &&
			ii->download && (ii->status != ItemInfo::STATUS_RUNNING)) {
			cd->clrText = WinUtil::blendColors(WinUtil::bgColor, WinUtil::textColor, 0.4);
			return CDRF_NEWFONT;
		} else if (ii->download && (ii->status != ItemInfo::STATUS_RUNNING)) {
			cd->clrText = WinUtil::textColor;
			return CDRF_NEWFONT;
		}
		// Fall through
	}
	default:
		return CDRF_DODEFAULT;
	}
}

LRESULT TransferView::onDoubleClickTransfers(int /*idCtrl*/, LPNMHDR pnmh, BOOL& /*bHandled*/) {
    NMITEMACTIVATE* item = (NMITEMACTIVATE*)pnmh;
	if(item->iItem != -1) {
		switch(SETTING(TRANSFERLIST_DBLCLICK)) {
			case 0:
				ctrlTransfers.getItemData(item->iItem)->pm();
				break;
			case 1:
				ctrlTransfers.getItemData(item->iItem)->getList();
				break;
			case 2:
				ctrlTransfers.getItemData(item->iItem)->matchQueue();
				break;
			case 4:
				ctrlTransfers.getItemData(item->iItem)->addFav();
				break;
		}
	}
	return 0;
}

int TransferView::ItemInfo::compareItems(ItemInfo* a, ItemInfo* b, int col) {
	if(BOOLSETTING(ALT_SORT_ORDER)) {
		if(a->download == b->download) {
			if(a->status != b->status) {
				return (a->status == ItemInfo::STATUS_RUNNING) ? -1 : 1;
			}
		} else {
			return a->download ? -1 : 1;
		}
	} else {
		if(a->status == b->status) {
			if(a->download != b->download) {
				return a->download ? -1 : 1;
			}
		} else {
			return (a->status == ItemInfo::STATUS_RUNNING) ? -1 : 1;
		}
	}
	
	switch(col) {
		case COLUMN_STATUS: return 0;
		case COLUMN_TIMELEFT: return compare(a->timeLeft, b->timeLeft);
		case COLUMN_SPEED: return compare(a->speed, b->speed);
		case COLUMN_SIZE: return compare(a->size, b->size);
		case COLUMN_RATIO: return compare(a->getRatio(), b->getRatio());
		default:
			return lstrcmpi(a->columns[col].c_str(), b->columns[col].c_str());
	}
}

LRESULT TransferView::onSpeaker(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/) {
	TaskQueue::List t;
	tasks.get(t);
	if(t.size() > 2) {
		ctrlTransfers.SetRedraw(FALSE);
	}

	for(TaskQueue::Iter i = t.begin(); i != t.end(); ++i) {	
		switch(i->first) {
			case ADD_ITEM: {
				auto_ptr<UpdateInfo> ui(reinterpret_cast<UpdateInfo*>(i->second));
				ItemInfo* ii = new ItemInfo(ui->user, ui->download);
				ii->update(*ui);
				ctrlTransfers.insertItem(ii, ii->download ? 0 : 1);
				break;
			}
			case REMOVE_ITEM: {
				auto_ptr<UpdateInfo> ui(reinterpret_cast<UpdateInfo*>(i->second));
				int ic = ctrlTransfers.GetItemCount(); 
				for(int i = 0; i < ic; ++i) {
					ItemInfo* ii = ctrlTransfers.getItemData(i);
					if(*ui == *ii) {
						ctrlTransfers.DeleteItem(i);
						delete ii;
						break;
					}
				}
				break;
			}
			case UPDATE_ITEM: {
				auto_ptr<UpdateInfo> ui(reinterpret_cast<UpdateInfo*>(i->second));
				int ic = ctrlTransfers.GetItemCount(); 
				for(int i = 0; i < ic; ++i) {
					ItemInfo* ii = ctrlTransfers.getItemData(i);
					if(ii->download == ui->download && ii->user == ui->user) {
						ii->update(*ui);
						ctrlTransfers.updateItem(i);						
						break;
					}
				}
				break;
			}
			default:
				break;
		}
	}

	if(t.empty() == false) {
		ctrlTransfers.resort();
		if(t.size() > 2) {
			ctrlTransfers.SetRedraw(TRUE);
		}
	}

	return 0;
}

TransferView::ItemInfo::ItemInfo(const User::Ptr& u, bool aDownload) : UserInfoBase(u), download(aDownload), transferFailed(false),
	status(STATUS_WAITING), pos(0), size(0), start(0), actual(0), speed(0), timeLeft(0), Target(Util::emptyString), flagImage(0)
{ 
	columns[COLUMN_USER] = WinUtil::getNicks(u);
	columns[COLUMN_HUB] = WinUtil::getHubNames(u).first;
}

void TransferView::ItemInfo::update(const UpdateInfo& ui) {
	if(ui.updateMask & UpdateInfo::MASK_STATUS) {
		status = ui.status;
	}
	if(ui.updateMask & UpdateInfo::MASK_STATUS_STRING) {
		// No slots etc from transfermanager better than disconnected from connectionmanager
		if(!transferFailed)
			columns[COLUMN_STATUS] = ui.statusString;
		transferFailed = ui.transferFailed;
	}
	if(ui.updateMask & UpdateInfo::MASK_SIZE) {
		size = ui.size;
		// gui cosmetic ;)
		if(size != -1) {
            columns[COLUMN_SIZE] = Text::toT(Util::formatBytes(size));
        } else {
            columns[COLUMN_SIZE] = Util::emptyStringT;
        }
	}
	if(ui.updateMask & UpdateInfo::MASK_START) {
		start = ui.start;
	}
	if(ui.updateMask & UpdateInfo::MASK_POS) {
		pos = start + ui.pos;
	}
	if(status == STATUS_RUNNING) {
    	if(ui.updateMask & UpdateInfo::MASK_SPEED) {
    		speed = ui.speed;
        	columns[COLUMN_SPEED] = Text::toT(Util::formatBytes(speed) + "/s");
    	}
    	if(ui.updateMask & UpdateInfo::MASK_TIMELEFT) {
    		timeLeft = ui.timeLeft;
    		columns[COLUMN_TIMELEFT] = Text::toT(Util::formatSeconds(timeLeft));
    	}
    	if(ui.updateMask & UpdateInfo::MASK_ACTUAL) {
    		actual = start + ui.actual;
    		columns[COLUMN_RATIO] = Text::toT(Util::toString(getRatio()));
    	}
    } else {
        if(ui.updateMask & UpdateInfo::MASK_SPEED) {
            columns[COLUMN_SPEED] = Util::emptyStringT;
        }
        if(ui.updateMask & UpdateInfo::MASK_TIMELEFT) {
            columns[COLUMN_TIMELEFT] = Util::emptyStringT;
        }
        if(ui.updateMask & UpdateInfo::MASK_ACTUAL) {
            columns[COLUMN_RATIO] = Util::emptyStringT;
        }
    }
	if(ui.updateMask & UpdateInfo::MASK_FILE) {
		if(ui.download) Target = Text::fromT(ui.target);
		columns[COLUMN_FILE] = ui.file;
		columns[COLUMN_PATH] = ui.path;
	}
	if(ui.updateMask & UpdateInfo::MASK_IP) {	
		flagImage = ui.flagImage;
		columns[COLUMN_IP] = ui.IP;
		IP = ui.sIP;
	}
}

void TransferView::on(ConnectionManagerListener::Added, ConnectionQueueItem* aCqi) {
	UpdateInfo* ui = new UpdateInfo(aCqi->getUser(), aCqi->getDownload());

	if(ui->download) {
		string aTarget; int64_t aSize;
		if(QueueManager::getInstance()->getQueueInfo(aCqi->getUser(), aTarget, aSize)) {
			ui->setFile(Text::toT(aTarget));
			ui->setSize(aSize);
		}
	}

	ui->setStatus(ItemInfo::STATUS_WAITING);
	ui->setStatusString(TSTRING(CONNECTING));

	speak(ADD_ITEM, ui);
}

void TransferView::on(ConnectionManagerListener::StatusChanged, ConnectionQueueItem* aCqi) {
	UpdateInfo* ui = new UpdateInfo(aCqi->getUser(), aCqi->getDownload());

	string aTarget;	int64_t aSize;
	if(QueueManager::getInstance()->getQueueInfo(aCqi->getUser(), aTarget, aSize)) {
		ui->setFile(Text::toT(aTarget));
		ui->setSize(aSize);
	}
	
	ui->setStatusString((aCqi->getState() == ConnectionQueueItem::CONNECTING) ? TSTRING(CONNECTING) : TSTRING(WAITING_TO_RETRY));

	speak(UPDATE_ITEM, ui);
}

void TransferView::on(ConnectionManagerListener::Removed, ConnectionQueueItem* aCqi) {
	speak(REMOVE_ITEM, new UpdateInfo(aCqi->getUser(), aCqi->getDownload()));
}

void TransferView::on(ConnectionManagerListener::Failed, ConnectionQueueItem* aCqi, const string& aReason) {
	UpdateInfo* ui = new UpdateInfo(aCqi->getUser(), aCqi->getDownload());
	ui->setStatusString(Text::toT(aReason));
	speak(UPDATE_ITEM, ui);
}

void TransferView::on(DownloadManagerListener::Starting, Download* aDownload) {
	UpdateInfo* ui = new UpdateInfo(aDownload->getUser(), true);
	ui->setStatus(ItemInfo::STATUS_RUNNING);
	ui->setPos(aDownload->getTotal());
	ui->setActual(aDownload->getActual());
	ui->setStart(aDownload->getPos());
	ui->setSize(aDownload->getSize());
	ui->setFile(Text::toT(aDownload->getTarget()));
	ui->setStatusString(TSTRING(DOWNLOAD_STARTING));
	tstring country = Text::toT(Util::getIpCountry(aDownload->getUserConnection().getRemoteIp()));
	tstring ip = Util::emptyStringT;
	OnlineUser& ou = ClientManager::getInstance()->getOnlineUser(ui->user);
	if(&ou && ou.getClient().isOp()) {
        ip = Text::toT(aDownload->getUserConnection().getRemoteIp());
        ui->sIP = Text::toT(aDownload->getUserConnection().getRemoteIp());
  }
			
	if(country.empty()) {
        ui->setIP(ip);
	} else {
        if(ip.empty()) {
            ui->setIP(country);
        } else {
            ui->setIP(country + _T(" (") + ip + _T(")"));
        }
	}
	
	ui->flagImage = WinUtil::getFlagImage(Text::fromT(country).c_str());
	
	if(aDownload->isSet(Download::FLAG_TREE_DOWNLOAD)) {
		ui->file = _T("TTH: ") + ui->file;
	}

	speak(UPDATE_ITEM, ui);
}

void TransferView::on(DownloadManagerListener::Tick, const Download::List& dl) {
	AutoArray<TCHAR> buf(TSTRING(DOWNLOADED_BYTES).size() + 64);

	for(Download::List::const_iterator j = dl.begin(); j != dl.end(); ++j) {
		Download* d = *j;

		UpdateInfo* ui = new UpdateInfo(d->getUser(), true);
		ui->setActual(d->getActual());
		ui->setPos(d->getTotal());
		ui->setTimeLeft(d->getSecondsLeft());
		ui->setSpeed(d->getRunningAverage());
		
		_stprintf_s(buf, TSTRING(DOWNLOADED_BYTES).size() + 64, CTSTRING(DOWNLOADED_BYTES), Text::toT(Util::formatBytes(d->getPos())).c_str(), 
			(double)d->getPos()*100.0/(double)d->getSize(), Text::toT(Util::formatSeconds((GET_TICK() - d->getStart())/1000)).c_str());

		tstring statusString;

		if(d->isSet(Download::FLAG_TTH_CHECK)) {
			statusString += _T("[T]");
		}
		if(d->isSet(Download::FLAG_ZDOWNLOAD)) {
			statusString += _T("[Z]");
		} 
		if(d->isSet(Download::FLAG_ROLLBACK)) {
			statusString += _T("[R]");
		}
		if(!statusString.empty()) {
			statusString += _T(" ");
		}
		statusString += buf;
		ui->setStatusString(statusString);

		tasks.add(UPDATE_ITEM, ui);
	}

	PostMessage(WM_SPEAKER);
}

void TransferView::on(DownloadManagerListener::Failed, Download* aDownload, const string& aReason) {
	UpdateInfo* ui = new UpdateInfo(aDownload->getUser(), true, true);
	ui->setStatus(ItemInfo::STATUS_WAITING);
	ui->setPos(0);
	ui->setStatusString(Text::toT(aReason));
	ui->setSize(aDownload->getSize());
	ui->setFile(Text::toT(aDownload->getTarget()));
	tstring country = Text::toT(Util::getIpCountry(aDownload->getUserConnection().getRemoteIp()));
	tstring ip = Util::emptyStringT;
	OnlineUser& ou = ClientManager::getInstance()->getOnlineUser(ui->user);
	if(&ou && ou.getClient().isOp()) {
        ip = Text::toT(aDownload->getUserConnection().getRemoteIp());
        ui->sIP = Text::toT(aDownload->getUserConnection().getRemoteIp());
  }
			
	if(country.empty()) {
        ui->setIP(ip);
	} else {
        if(ip.empty()) {
            ui->setIP(country);
        } else {
            ui->setIP(country + _T(" (") + ip + _T(")"));
        }
	}
	
	ui->flagImage = WinUtil::getFlagImage(Text::fromT(country).c_str());
	
	if(aDownload->isSet(Download::FLAG_TREE_DOWNLOAD)) {
		ui->file = _T("TTH: ") + ui->file;
	}

	speak(UPDATE_ITEM, ui);
}

void TransferView::on(UploadManagerListener::Starting, Upload* aUpload) {
	UpdateInfo* ui = new UpdateInfo(aUpload->getUser(), false);

	ui->setStatus(ItemInfo::STATUS_RUNNING);
	ui->setPos(aUpload->getTotal());
	ui->setActual(aUpload->getActual());
	ui->setStart(aUpload->getPos());
	ui->setSize(aUpload->getSize());
	ui->setFile(Text::toT(aUpload->getSourceFile()));
	ui->setStatusString(TSTRING(UPLOAD_STARTING));
	tstring country = Text::toT(Util::getIpCountry(aUpload->getUserConnection().getRemoteIp()));
	tstring ip = Util::emptyStringT;
	OnlineUser& ou = ClientManager::getInstance()->getOnlineUser(ui->user);
	if(&ou && ou.getClient().isOp()) {
        ip = Text::toT(aUpload->getUserConnection().getRemoteIp());
        ui->sIP = Text::toT(aUpload->getUserConnection().getRemoteIp());
    }
			
	if(country.empty()) {
        ui->setIP(ip);
	} else {
        if(ip.empty()) {
            ui->setIP(country);
        } else {
            ui->setIP(country + _T(" (") + ip + _T(")"));
        }
	}
	
	ui->flagImage = WinUtil::getFlagImage(Text::fromT(country).c_str());
	if(aUpload->isSet(Upload::FLAG_TTH_LEAVES)) {
		ui->file = _T("TTH: ") + ui->file;
	}

	speak(UPDATE_ITEM, ui);
}

void TransferView::on(UploadManagerListener::Tick, const Upload::List& ul) {
	AutoArray<TCHAR> buf(TSTRING(UPLOADED_BYTES).size() + 64);

	for(Upload::List::const_iterator j = ul.begin(); j != ul.end(); ++j) {
		Upload* u = *j;

		UpdateInfo* ui = new UpdateInfo(u->getUser(), false);
		ui->setActual(u->getActual());
		ui->setPos(u->getTotal());
		ui->setTimeLeft(u->getSecondsLeft());
		ui->setSpeed(u->getRunningAverage());

		_stprintf_s(buf, STRING(UPLOADED_BYTES).size() + 64, CTSTRING(UPLOADED_BYTES), Text::toT(Util::formatBytes(u->getPos())).c_str(), 
			(double)u->getPos()*100.0/(double)u->getSize(), Text::toT(Util::formatSeconds((GET_TICK() - u->getStart())/1000)).c_str());

		tstring statusString;
			
		if(u->isSet(Upload::FLAG_ZUPLOAD)) {
			statusString = _T("[Z] ") + tstring(buf);
		} else {
			statusString = buf;
		}

		ui->setStatusString(statusString);

		tasks.add(UPDATE_ITEM, ui);
	}

	PostMessage(WM_SPEAKER);
}

void TransferView::onTransferComplete(Transfer* aTransfer, bool isUpload) {
	UpdateInfo* ui = new UpdateInfo(aTransfer->getUser(), !isUpload);

	ui->setStatus(ItemInfo::STATUS_WAITING);
	ui->setPos(0);
	ui->setStatusString(isUpload ? TSTRING(UPLOAD_FINISHED_IDLE) : TSTRING(DOWNLOAD_FINISHED_IDLE));

	speak(UPDATE_ITEM, ui);
}

void TransferView::ItemInfo::disconnect() {
	ConnectionManager::getInstance()->disconnect(user, download);
}

LRESULT TransferView::onSearchAlternates(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(ctrlTransfers.GetSelectedCount() == 1) {
		int i = ctrlTransfers.GetNextItem(-1, LVNI_SELECTED);
		ItemInfo* ii = ctrlTransfers.getItemData(i);

		string target = ii->Target;

		TTHValue tth;
		if(QueueManager::getInstance()->getTTH(target, tth)) {
			WinUtil::searchHash(&tth);
		}
	} 
	return 0;
}

LRESULT TransferView::onCopyIP(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	int i = -1;
	tstring IP;
	while( (i = ctrlTransfers.GetNextItem(i, LVNI_SELECTED)) != -1) {
		IP = ctrlTransfers.getItemData(i)->IP;
		WinUtil::setClipboard(IP);
	}
	return 0;
}

LRESULT TransferView::onPreviewCommand(WORD /*wNotifyCode*/, WORD wID, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
    if(ctrlTransfers.GetSelectedCount() == 1) {
    	string sTempTarget = QueueManager::getInstance()->getTempTarget(ctrlTransfers.getItemData(ctrlTransfers.GetNextItem(-1, LVNI_SELECTED))->Target);
    	if(sTempTarget.empty() == false) {
    		WinUtil::RunPreviewCommand(wID - IDC_PREVIEW_APP, sTempTarget);
    	}
    }
	return 0;
}

LRESULT TransferView::onWhoisIP(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	int i = -1;
	tstring IP;
	while( (i = ctrlTransfers.GetNextItem(i, LVNI_SELECTED)) != -1) {
		IP = ctrlTransfers.getItemData(i)->IP;
		if(IP != _T(""))
			WinUtil::openLink(_T("http://www.ripe.net/perl/whois?form_type=simple&full_query_string=&searchtext=") + IP);
	}
	return 0;
}

void TransferView::on(SettingsManagerListener::Save, SimpleXML& /*xml*/) throw() {
    PostMessage(WM_REFRESH_SETTINGS);
}

void TransferView::on(DownloadManagerListener::HashCheck, Download* aDownload) {
	UpdateInfo* ui = new UpdateInfo(aDownload->getUser(), true);

	ui->setStatusString(TSTRING(CHECKING_TTH));

	speak(UPDATE_ITEM, ui);
}

LRESULT TransferView::onOpenUserLog(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
    ItemInfo *ii = ctrlTransfers.getItemData(ctrlTransfers.GetNextItem(-1, LVNI_SELECTED));
	StringMap params;
	params["userNI"] = ii->user->getFirstNick();
	params["userCID"] = ii->user->getCID().toBase32();
	params["hubNI"] = Util::toString(ClientManager::getInstance()->getHubNames(ii->user->getCID()));
	params["hubURL"] = Util::toString(ClientManager::getInstance()->getHubs(ii->user->getCID()));
	params["myCID"] = ClientManager::getInstance()->getMe()->getCID().toBase32();
	string file = Util::validateFileName(SETTING(LOG_DIRECTORY) + Util::formatParams(SETTING(LOG_FILE_PRIVATE_CHAT), params, true));
	if(File::existsFile(file)) {
		ShellExecute(NULL, NULL, Text::toT(file).c_str(), NULL, NULL, SW_SHOWNORMAL);
	} else {
		MessageBox(CTSTRING(NO_LOG_FOR_USER),CTSTRING(NO_LOG_FOR_USER), MB_OK );	  
	}
	return 0;
}

LRESULT TransferView::onRefreshSettings(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/) {
    bool refresh = false;
    if(ctrlTransfers.GetBkColor() != WinUtil::bgColor) {
        ctrlTransfers.SetBkColor(WinUtil::bgColor);
        ctrlTransfers.SetTextBkColor(WinUtil::bgColor);
        refresh = true;
    }
    if(ctrlTransfers.GetTextColor() != WinUtil::textColor) {
        ctrlTransfers.SetTextColor(WinUtil::textColor);
        refresh = true;
    }
    if(refresh == true) {
        RedrawWindow(NULL, NULL, RDW_ERASE | RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN);
    }
	return 0;
}

/* Some parts Copyright (C) 2002-2008 Petr Kozelka, PPK at czdc dot org */
